uses CRT, DOS, RB_unit, CBM_sw1;

const
bitBiDir        = $20;

bitErrorOut	= $01;
bitCmdOut 	= $02;
bitDataOut 	= $04;
bitBusyOut	= $08;
bitErrorIn	= $08;
bitCmdIn 	= $20;
bitDataIn 	= $40;
bitBusyIn	= $80;

coProgName      = '1541LPT';

H = True;
L = False;



var
  bHandle1, bHandle2, bLPTctrl, bSector : byte;
  iTrack : integer;
  sDrive, sPort : Str23;
  sImage : string;


{============================================================================}


function fBusyIn : boolean;
var
  b : byte;
begin
{
  repeat
   b := port[wIeeeLPTstat];
  until (b = port[wIeeeLPTstat]);

  fBusyIn := (b and bitBusyIn > 0);
}
  fBusyIn := (port[wIeeeLPTstat] and bitBusyIn > 0);
end;


function fCmdIn : boolean;
var
  b : byte;
begin
{
  repeat
   b := port[wIeeeLPTstat];
  until (b = port[wIeeeLPTstat]);

  fCmdIn := (b and bitCmdIn = 0);
}
  fCmdIn := (port[wIeeeLPTstat] and bitCmdIn = 0);
end;


function fDataIn : boolean;
var
  b : byte;
begin
{
  repeat
   b := port[wIeeeLPTstat];
  until (b = port[wIeeeLPTstat]);

  fDataIn := (b and bitDataIn = 0);
}
  fDataIn := (port[wIeeeLPTstat] and bitDataIn = 0);
end;


function fErrorIn : boolean;
var
  b : byte;
begin
{
  repeat
   b := port[wIeeeLPTstat];
  until (b = port[wIeeeLPTstat]);

  fErrorIn := (b and bitErrorIn = 0);
}
  fErrorIn := (port[wIeeeLPTstat] and bitErrorIn = 0);
end;


{============================================================================}


procedure ParallelInput;
{-----------------------------------------------------------------------------
Switch the parallel port to input
-----------------------------------------------------------------------------}
begin
  port[wIeeeLPTaddr] := $FF;
  bLPTctrl := bLPTctrl or bitBidir;
  port[wIeeeLPTctrl] := bLPTctrl;
end;


procedure ParallelOutput;
{---------------------------------------------------------------------------
Switch the parallel port of the hybrid/parallel cable to output
---------------------------------------------------------------------------}
begin
  bLPTctrl := bLPTctrl and (not bitBidir);
  port[wIeeeLPTctrl] := bLPTctrl;
end;


procedure pBusyOut(b : boolean);
begin
  if b then bLPTctrl := bLPTctrl and not(bitBusyOut)
   else bLPTctrl := bLPTctrl or bitBusyOut;
  port[wIeeeLPTctrl] := bLPTctrl;
end;


procedure pCmdOut(b : boolean);
begin
  if b then bLPTctrl := bLPTctrl and not(bitCmdOut)
   else bLPTctrl := bLPTctrl or bitCmdOut;
  port[wIeeeLPTctrl] := bLPTctrl;
end;


procedure pDataOut(b : boolean);
begin
  if b then bLPTctrl := bLPTctrl or bitDataOut
   else bLPTctrl := bLPTctrl and not(bitDataOut);
  port[wIeeeLPTctrl] := bLPTctrl;
end;


procedure pErrorOut(b : boolean);
begin
  if b then bLPTctrl := bLPTctrl and not(bitErrorOut)
   else bLPTctrl := bLPTctrl or bitErrorOut;
  port[wIeeeLPTctrl] := bLPTctrl;
end;


{============================================================================}


function ByteIn : byte;
var
  by : byte;
begin
  repeat
   write;
  until fDataIn;

  repeat
   by := port[wIeeeLPTaddr];
  until by = port[wIeeeLPTaddr];

  pDataout(L);

  repeat
   write;
  until not fDataIn;

  pDataout(H);
  ByteIn := by;
end;


function zin(w:word):string;
begin
  case w of
    0 : zin := coProgName;
    1 : zin := 'Floppy simulator for Commodores with IEC-bus';
    2 : zin := 'Copyright (c) 1989-2005 by Ruud Baltissen';
    3 : zin := 'Connection with 1541 is being tested....';
    4 : zin := 'Connection with 1541 is OK!';
    5 : zin := ', INI-file does not exist !!!';
    6 : zin := 'Wrong PORT parameter !!!';
    7 : zin := 'Wrong LPT port number !!!';
    8 : zin := 'Use /? for help';
    9 : zin := 'No LPT port detected !!!';
   10 : zin := 'Wrong DRIVE parameter !!!';
     else zin := '???';
  end;
end;


{============================================================================}


procedure ByteOut(by : byte);
begin
  repeat
   write;
  until not fDataIn;

  ParallelOutput;
  port[wIeeeLPTaddr] := by;
  pDataOut(L);

  repeat
   write;
  until fDataIn;

  pDataOut(H);

  repeat
   write;
  until not fDataIn;

  ParallelInput;
end;


procedure HelpMessage;
begin
end;


procedure ReadParameters;
{---------------------------------------------------------------------------
Reads parameters behind the command and checks if they are valid
 in:  -
 out: -
---------------------------------------------------------------------------}
var
  pbo          : boolean;
  by           : byte;
  I,i2         : integer;
  plo          : longint;
  Cmd,pst,pst2 : string;
begin
  pbo      := false;


{ Check number of parameters }
  if paramcount = 0 then HelpMessage;

{ Check parameters one by one }
  for I := 1 to ParamCount do
   begin

    Cmd := Upper(ParamStr(I));

    if Copy(Cmd, 1, 2) = '/D' then
     begin

  { type drive to emulate }
      sDrive := copy(Cmd, 3, length(Cmd)-2);
     end
      else

    if Copy(Cmd, 1, 2) = '/F' then
     begin

  { Parameter is name of image}
      sImage := copy(Cmd, 3, length(Cmd)-2);
     end
      else

    if Copy(Cmd, 1, 2) = '/P' then
     begin

  { type drive to emulate }
      sPort := copy(Cmd, 3, length(Cmd)-2);
     end
      else

    if (Cmd = '/?') Or (Cmd = '/H') then
     begin

  { Helpmessage }
      HelpMessage;
     end
      else

     pbo := true;
   end;

{ (PBO = true) = Error! }
  if (pbo or (sPort = '') or (sImage = '') or (sDrive = '')) then HelpMessage;
end;


{============================================================================}


procedure main;
var
  boDir0, boDir1, boDirImages : boolean;
  by, by2, by3, bt1, bt2, bChannel : byte;
  I, iError : integer;
  DL0, DL1 : pDirLine;
  DirInfo : SearchRec;
  pst, sBAMlabel : Str23;
  Emu : TEmutype;
  wt, wtp : word;
label PL1, PL2;
begin
  if IsVGA then textmode(CO80 + font8x8);
  MaxX := 80;
  writeln;
  for wt := 1 to 2 do MWriteln(zin(wt), MaxX);
  writeln;

  wtp := 1;
  iTrack := 18;
  bSector := 0;
  new(Device[1]);
  new(Device[2]);

  bHandle1 := AssignHandle(0, 'Dev1_Chan0');
  Device[1]^.Channel[bChannel].cHandle := bHandle1;
  DriveHandles[bHandle1].hChannel := bChannel;

  bHandle2 := AssignHandle(0, 'Dev2_Chan0');
  Device[2]^.Channel[bChannel].cHandle := bHandle2;
  DriveHandles[bHandle2].hChannel := bChannel;

  ReadParameters;

{**  Validate read parameters  **}
{ 'Disable' all ports }
  wIeeeComAddr := 0;
  wIeeeLptAddr := 0;
  wIeeeLptCtrl := 0;
  wIeeeLptStat := 0;

{ Find first available LPT port }
  For by := 1 to 4 Do
   begin
    wIeeeLptAddr := BIOSLPT(by);
    if wIeeeLptAddr <> 0 then Break;
   end;

  if wIeeeLptAddr = 0 then
   begin
    Writeln(zin(9));
    Halt(1);
   end;

  with DriveHandles[bHandle1] do
   begin
    hFilename := sImage;
    hbDevNo := 1;
    hbDrvNo := 0;
    Emu := EmuInActive;

{ What LPT-port is used? }
    val(sPort, I, IError);
    if ((IError <> 0) or (I < 0) or (I > 3)) then
     begin
      writeln(zin(7));
      Halt(1);
     end;

    wIeeeLptAddr := BIOSLPT(I);

    if (wIeeeLptAddr = 0) then
     begin
      Writeln(zin(9));
      Halt(1);
     end;

    wIeeeLptStat := wIeeeLptAddr + 1;
    wIeeeLptCtrl := wIeeeLptAddr + 2;
    writeln('LPT ', dec2hex(wIeeeLptAddr, 3));

    writeln;
    writeln;

{ What drive is used? }
    if sDrive = '1541' then Emu := Emu1541 else
    if sDrive = '1571' then Emu := Emu1571 else
    if sDrive = '1581' then Emu := Emu1581 else
    if sDrive = '3040' then Emu := Emu8050 else
    if sDrive = '4040' then Emu := Emu4040 else
    if sDrive = '8050' then Emu := Emu8050 else
    if sDrive = '8250' then Emu := Emu8250 else
    if sDrive = 'CBM'  then Emu := EmuCBM  else
    if sDrive = 'D16'  then Emu := EmuD16  else
    if sDrive = 'PC'   then Emu := EmuPC;

{#### Skip devices not supported yet ####}
if (Emu <> Emu1541) then Emu := EmuInActive;

    Fillchar(Device[hbDevNo]^, Sizeof(TEmuDevice), 0);
    Device[hbDevNo]^.bLastDrive := 0;
    Device[hbDevNo]^.FSType := Emu;
    Device[hbDevNo]^.wDirSize := 0;
    Device[hbDevNo]^.wDirPtr := 0;

    for by := 0 to 1 do Device[hbDevNo]^.Drv[by].Inuse := False;

    for by := 0 to 15 do
     begin
      Device[hbDevNo]^.BufFree[by] := True;

      with Device[hbDevNo]^.Channel[by] do
       begin
        cHandle := 0;
        Opened := False;
        Command := '';
       end;
     end;

    Device[hbDevNo]^.FSType := Emu;
    Device[hbDevNo]^.dErrorMsg := '73, 1541LPT V1.0, 00, 00'#13;

    with Device[hbDevNo]^.Drv[hbDrvNo] do
     begin
      PCfile := sImage;
      LastTrack := 0;
      InUse := True;
      sDirPath := chCsep;
     end;

    if Emu in [EmuCBM, EmuPC] then
     begin
      Device[hbDevNo]^.Drv[hbDrvNo].DosDrive := ord(sImage[1]) - 64;
      Device[hbDevNo]^.Drv[hbDrvNo].sDirPath := sImage + chCsep;
     end
    else Device[hbDevNo]^.Drv[hbDrvNo].DosDrive := $FF;

    if (not exists(sImage)) then
     CreateDiskImage(0, '1541LPT', 'RB ' + DiskIDcode(bHandle1));
   end;

  with DriveHandles[2] do
   begin
    hFilename := 'DIRIMAGE.D64';
    hbDevNo := 2;
    hbDrvNo := 0;
    hDrvNum := 1;
    hdTrack := DIRtrack(1);
    hdSector := DIRsector(1);

    Fillchar(Device[hbDevNo]^, Sizeof(TEmuDevice), 0);
    Device[hbDevNo]^.bLastDrive := 0;
    Device[hbDevNo]^.FSType := Emu;
    Device[hbDevNo]^.wDirSize := 0;
    Device[hbDevNo]^.wDirPtr := 0;

    for by := 0 to 1 do Device[hbDevNo]^.Drv[by].Inuse := False;

    for by := 0 to 15 do
     begin
      Device[hbDevNo]^.BufFree[by] := True;

      with Device[hbDevNo]^.Channel[by] do
       begin
        cHandle := 0;
        Opened := False;
        Command := '';
       end;
     end;

    Device[hbDevNo]^.FSType := Emu;
    Device[hbDevNo]^.dErrorMsg := '73, 1541LPT V1.0, 00, 00'#13;

    with Device[hbDevNo]^.Drv[hbDrvNo] do
     begin
      PCfile := 'DIRIMAGE.D64';
      LastTrack := 0;
      InUse := True;
      sDirPath := chCsep;
      dTrack := DIRtrack(1);
      dSector := DIRsector(1);
     end;

    Device[hbDevNo]^.Drv[hbDrvNo].DosDrive := $FF;
   end;


PL1:
  boDirImages := False;
  bLPTctrl := $FF;
  ParallelInput;
  pBusyOut(L);
  pCmdOut(H);
  pDataOut(H);
  pErrorOut(H);

  bChannel := 0;

  InitDxx(bHandle1);
  with DriveHandles[bHandle1] do
   sBAMlabel := Device[hbDevNo]^.Drv[hbDrvNo].BAMlabel;


{**  Check communication with 1541  **}
  pCmdOut(L);
  pDataOut(L);
  pErrorOut(L);

  writeln;
  writeln('Waiting for connection....');

  repeat
   if keypressed then halt(1);
  until (not fBusyIn) and fErrorIn and (not fDataIn) and fCmdIn;

  pCmdOut(H);
  pDataOut(H);
  pErrorOut(H);

  ParallelOutput;
  port[wIeeeLPTaddr] := $7E;
  pBusyOut(H);

  repeat until (not fBusyIn) and (not fErrorIn) and
               (not fDataIn) and (not fCmdIn);

  ParallelInput;

  repeat until fCmdIn;
  pCmdOut(L);
  repeat until not fCmdIn;
  pCmdOut(H);

  repeat until fDataIn;
  pDataOut(L);
  repeat until not fDataIn;
  pDataOut(H);

  repeat until fErrorIn;
  pErrorOut(L);
  repeat until not fErrorIn;
  pErrorOut(H);

  repeat until fBusyIn;
  pBusyOut(L);
  repeat until not fBusyIn;

  pBusyOut(H);
  pCmdOut(H);
  pDataOut(H);
  pErrorOut(H);

  for by := 0 to 255 do ByteOut(by);

  for by := 255 downto 0 do
   begin
    pBusyOut(H);
    by := ByteIn;
   end;

  pBusyOut(H);
  pCmdOut(H);
  pDataOut(H);
  pErrorOut(H);

  writeln(zin(4));
  writeln;

  repeat
{   Write('Waiting for command... ');}

   repeat
    if (not fBusyIn) and fErrorIn and (not fDataIn) and fCmdIn then
     goto PL1;
   until (fCmdIn and (not fErrorIn)) or KeyPressed;

   if KeyPressed then
    begin
     while Keypressed do Readkey;
     halt(0);
    end;

   pBusyOut(H);
   by := ByteIn;
   if not(by in [36, 37]) then write(by:3, '  ');

   if by = 1 then
    begin

{**  Set track and sector  **}
     iTrack := ByteIn;
     bSector := ByteIn;

     pBusyOut(L);
     Writeln('Setting track and sector ', iTrack, ' ', bSector);
     pBusyOut(H);
    end
     else

   if by = 2 then
    begin
     writeln('Verification ', ByteIn);
    end
     else

   if (by in [3, 9]) then
    begin

{**  Send sector  **}
     pBusyOut(L);

     if (by = 3) then Write('Sending sector.... ') else
      Write('Verifying sector.... ');

     if boDirImages then
      begin
       if    (iTrack <> DIRtrack(8))
          or ((bSector = 0) and boDir0)
          or ((bSector = 1) and boDir1) then
        begin
         boDirImages := False;
         by2 := bHandle1;
        end
       else
        begin
         if (bSector = 0) then
          begin
           by2 := bHandle1;
           boDir0 := True;
          end
         else by2 := bHandle2;

         boDir1 := (bSector = 1);
        end;

       with DriveHandles[by2] do
        Device[hbDevNo]^.Drv[hbDrvNo].LastTrack := 0;
      end
     else by2 := bHandle1;

     ReadSector(by2, iTrack, bSector);

     if boDirImages then
      begin
       if (Device[2]^.Drv[0].dDataBlock[0] = 0) then
        begin
         boDirImages := False;

       { make sure drive rereads BAM }
         with DriveHandles[1] do
          Device[hbDevNo]^.Drv[hbDrvNo].LastTrack := 0;
        end;

       Device[1]^.Drv[0].dDataBlock := Device[2]^.Drv[0].dDataBlock;
      end;

     by3 := 0;
     pBusyOut(H);

     for wt := 0 to 255 do
      begin
       by2 := Device[1]^.Drv[0].dDataBlock[wt];
       ByteOut(by2);
       by3 := by3 xor by2;
      end;

     ByteOut(by3);                              { checksum }
     Writeln(' done');
    end
     else

   if by = 4 then
    begin

{**  Write sector  **}
     Write('Receiving sector.... ');
     for wt := 0 to 255 do Device[1]^.Drv[0].dDataBlock[wt] := ByteIn;
     pBusyOut(L);
     WriteSector(bHandle1, iTrack, bSector);

     if (iTrack = 18) and (bSector = 0) then
      begin
       InitDxx(bHandle1);
       with DriveHandles[0] do
        sBAMlabel := Device[hbDevNo]^.Drv[hbDrvNo].BAMlabel;

       write(sBAMlabel, ' ');
      end;

     Writeln(' done');
     pBusyOut(H);
    end
     else

   if by = 5 then
    begin
{**  Step head  **}
     by := ByteIn;
     pBusyOut(L);

     if (by > 127) then
      begin
       by := (256 - by) div 2;
       writeln('Move head ', by, ' tracks outwards');
       iTrack := iTrack - by;
       if (iTrack < 1) then iTrack := 1;
      end
     else
      begin
       by := by div 2;
       writeln('Move head ', by, ' tracks inwards');
       iTrack := iTrack + by;
       if (iTrack > 40) then iTrack := 40;
      end;

     pBusyOut(H);
    end
     else

   if by = 6 then
    begin

{**  Format track  **}
     iTrack := ByteIn;
     sBAMlabel[19] := chr(ByteIn);
     sBAMlabel[20] := chr(ByteIn);
     pBusyOut(L);
     Writeln('Formatting track ', iTrack, '....');
     for wt := 0 to 255 do Device[1]^.Drv[0].dDataBlock[wt] := 1;
     for wt := 0 to MaxSectors(bHandle1, by) do
      WriteSector(bHandle1, iTrack, wt);
     pBusyOut(H);
    end
     else

   if by = 7 then
    begin
     write('Read header: ', chr(ByteIn), ' ', chr(ByteIn), ' ');
     write(ByteIn, ' ', ByteIn, '   JOBN: ', ByteIn, ' ');

     for wt := 0 to 5 do
      begin
       write(Dec2Hex(ByteIn, 2), '  ');
      end;

     writeln;
    end
     else

   if by = 8 then
    begin
     writeln('Send ID: ', sBAMlabel[19], sBAMlabel[20]);
     ByteOut(ord(sBAMlabel[19]));
     ByteOut(ord(sBAMlabel[20]));
    end
     else

   if by = 10 then
    begin

{**  Initialize  **}
     by2 := ByteIn;                             { number of chars }
     write('Initialize ', by2, '  ');
     pst := '';
     for by3 := 1 to by2 do pst := pst + chr(ByteIn);
     writeln(pst);
     by3 := pos(':', pst);

     if (by3 = 0) or (by3 = length(pst)) then ByteOut(0) else
      begin
       delete(pst, 1, by3);
       pst := pst + '.D64';

       if exists(pst) then
        begin
         with Device[1]^.Drv[0] do
          begin
           PCfile := pst;
           LastTrack := 0;
           sDirPath := chCsep;
          end;

         InitDxx(bHandle1);
         with DriveHandles[0] do
          sBAMlabel := Device[1]^.Drv[0].BAMlabel;
         ByteOut(0);
        end
       else ByteOut(1);
      end;
    end
     else

   if by = 11 then
    begin

{**  Send directory of available images  **}
PL2:
     pBusyOut(L);
     writeln('Directory of available images...');
     boDirImages := True;
     boDir0 := False;
     boDir1 := False;
     pst := DxxExtension(bHandle1);
     delete(pst, 1, 1);
     CreateDiskImage(bHandle2, pst + ' IMAGE', 'RB ' + DiskIDcode(bHandle1));

     FindFirst('*' + DxxExtension(bHandle1), $27, DirInfo);

     while (DosError = 0) do
      begin

       with DriveHandles[bHandle2] do with Device[hbDevNo]^.Drv[hbDrvNo] do
        begin
         hFileType := $82;
         hFileName := upper(DirInfo.Name);
        end;

       AddDirEntry(bHandle2);
       FindNext(DirInfo);
      end;

     pBusyOut(H);
    end
     else

   if by = $F0 then Break;
  until False;

  dispose(Device[2]);
  dispose(Device[1]);
end;

begin
  Main;
end.



(*

 ; control/handshake
 ; bit 0 <- strobe    error
 ; bit 1 <- autofeed  command
 ; bit 2 -> error     error
 ; bit 3 <- initial.  data
 ; bit 4 <- select    busy
 ; bit 5 -> acknowl.  data
 ; bit 6 -> busy      busy	(H)
 ; bit 7 -> paper     command



3BC,378 = Printerdata

3BD,379 = Printerstatus
 bit 0 = not used
 bit 1 = not used
 bit 2 = not used
 bit 3 = error (15)
 bit 4 = select (13)
 bit 5 = paper end (12)
 bit 6 = acknowledge (10)
 bit 7 = busy (11)            - inverted

3BE,37A = Printercontrol
 bit 0 = strobe (1)           - inverted
 bit 1 = autofeed (14)        - inverted
 bit 2 = Initialise (16)
 bit 3 = select in (17)       - inverted
 bit 4 = enable IRQ7
 bit 5 = bi-directional mode
 bit 6 = not used
 bit 7 = not used

Control after reset:  $CC


*)