unit CBM_SW1;

interface


const
  chCsep : char = '\';
  chSep  : char = '\';

  coCheckIEEE       = 1000;
  coMaxDevice       = 31;
  coMaxDriveHandles = 63;
  coTimeOut         = 20000;

  Power2 : array[0..7] of byte = (1, 2, 4, 8, 16, 32, 64, 128);

  sbBidir        = $20;
  ssEOF          = $40;
  ssNoDevice     = $80;

  stCBMdir       = 'CBM-DIR.TXT';


type
  Block256 = array[0..255] of byte;
  Str2 = string[2];
  Str16 = string[16];
  Str32 = string[32];
  Str80 = string[80];
  Str23 = string[23];
  TADir = array[0..299] of Str80;

  TGCRbytes = record
    g1, g2, g3, g4, g5 : byte;
   end;


  TEmutype = (Emu1541,
              Emu1571,
              Emu1581,
              Emu8050,
              Emu3040,
              Emu4040,
              Emu8250,
              EmuCBM,                   { MSDOS but the in C= format }
              EmuD16,
              EmuInActive,
              EmuP00,
              EmuLPT,
              EmuPC,                    { MSDOS }
              EmuCOM,
              EmuT64);

  TAccess  = (fREAD, fWRITE, fMODIFY);

  TMODE = (tmNone, tmTalk, tmListen);

  pDirLine = ^tDirLine;

  tDirLine = record
    Size : longint;
    str : Str80;
    nLine : pDirLine;
   end;

  TBAMInfo = record
    FreeCount : Byte;
    BitTrack  : Array [0..4] of Byte;
   end;

  TBAM16Info = Array [0..31] of Byte;

  TBAMpos = record
    Track,
    Sector,
    bByt,
    bBit   : byte;
   end;

  TEmuDevice = record
    bLastDrive : byte;                  { last used drive }
    FStype     : TEmutype;
    wDirPtr,
    wDirSize: word;
    dErrorMsg  : Str80;

    Drv : array[0..1] of record
      sDirPath,                         { DOS path }
      PCfile     : Str80;               { name of Dxx file }
      dNoAccess,
      InUse      : boolean;
      DosDrive,
      LastTrack,                        { last accessed track }
      LastSector,                       { last accessed sector }
      dTrack,                           { first track of (sub) directory }
      dSector,                          { first sector of (sub) directory }
      dSize      : byte;
      DiskStatus : Integer;             { 0 = OK, rest is wrong }
      BAMlabel   : Str23;
      dDatablock : Block256;
     end;

    Channel : Array [0..15] of record
      Opened  : Boolean;
      BufPntr,
      PBuffer : byte;
      cHandle : integer;
      Command : str80;
     end;

    BufFree  : array[0..15] of boolean;
    Memory   : array[0..$0FFF] of byte;
   end;

  PDirEntryDxx = ^TDirEntryDxx;
  TDirEntryDxx = record
    Filetype,
    Track_1,
    Sector_1   : Byte;
    FName      : Array[0..15] of Char;
    TSideSec,
    SSideSec,
    RecLength  : byte;
    Unused     : Array[0..3] of Byte;
    Track_New,
    Sector_New : Byte;
    Lo_Size,
    Hi_Size    : Byte;
   end;

  TDriveHandle = record
    hFilename   : str80;
    hDrvNum,
    hbDevNo,                            { device where file is may be found }
    hbDrvNo,                            { drive where file is may be found }
    hSlot,
    hdTrack,                            { track of directory entry }
    hdSector,                           { sector of directory entry }
    hTrack,                             { Current track }
    hSector,                            { Current sector }
    hChannel,                           { current channel }
    hsTrack,                            { track of sidesectors }
    hsSector,                           { sector of sidesectors }
    hFiletype,                          { C= filetype / DOS-attribute }
    hBlkPnt     : byte;                 { pointer for hBlock }
    hOffset     : Integer;              { Offset in Sector }
    hPosition   : Longint;              { Position in file in bytes }
    hRecord,                            { recordpointer REL-file }
    hRecLen,                            { size of record in REL-file }
    hBlockSize  : Word;                 { Size }
    hFileFound,
    hForceDel,                          { delete file, locked or not }
    hReplace    : boolean;
    HAccess     : TAccess;
    hBlock      : Block256;             { memory for sequential operations }
   end;

  TProc = procedure;


var
  DriveHandles  : array[0..coMaxDriveHandles] of TDriveHandle;
  Device        : array[0..coMaxDevice] of ^TEmuDevice;

  boATNfound,
  boEOI,
  boError,
  boTest,
  boTestEnd     : boolean;

  Channel,
  ChannelKind,
  bByteIn,
  bByteInCnt1,
  bByteInCnt2,
  bIeeeType,            { 0 = internal, bi-directional port }
                        { 1 = external, bi-directional port }
                        { 2 = external, epp/ecp port }
                        { 5 = two LPT ports, at least one bi-directional }
  bOldError,
  bSerial,
  bSerial2,
  bStatus,
  MaxX          : byte;

  Error         : integer;                      { IO transfer error }

  ActiveDevs    : Set of Byte;

  sError        : Str32;

  OrgDir,
  TempDir       : Str80;

  AtnLo,
  AtnHi,
  ClkLo,
  ClkHi,
  DataLo,
  DataHi        : TProc;


  wIeeeComAddr, { base address COM port / control second LPT port }
  wIeeeLptAddr, { data LPT port }
  wIeeeLptCtrl, { control LPT port }
  wIeeeLptStat, { status LPT port }
  GTracksize,   { Size of GCR track }

                { Timing bus    Description      min.   Typ.    Max.  }
  Tne,          { Non-EOI response to RFD          -     40s   200s }
  Ts,           { Bit set-up talker               20s   70s     -   }
  Tv,           { Data valid                      20s   20s     -   }
  Tf,           { Frame handshake                  0s   20s  1000s }
  Ta,           { Frame to release of ATN         20s    -       -   }
  Tbb,          { Between bytes time             100s    -       -   }
  Try,          { Talker response limit            0s   30s    60s }
  Ttk,          { Talk-attention release          20s   30s   100s }
  Tdc  : Word;  { EOI acknoledge                  60s    -       -   }



{
trb : word;
tBuf : array[0..9] of byte;
DebugR : record
 ThbDrvNo, ThbDevno, ThSlot, ThdSector, ThdTrack, TdSector, TdTrack
  , ThTrack, ThSector : byte;
 tErrMsg,
 tPCFile : str80;
 Tdatablok : Block256;
end;
}



function  AtnFound : boolean;
function  DavFound : boolean;
function  EoiFound : boolean;
function  NdacFound : boolean;
function  NrfdFound : boolean;
function  ResetFound : boolean;

function  AddDirEntry(H : Integer) : boolean;
function  AllocateSector(H : integer) : boolean;
function  AssignHandle(H : integer; FName : Str16) : Integer;
function  ByteIn(H : integer; Var EOI : Boolean) : Byte;
function  CreateDiskImage(H : integer; stLabel, stID : str16) : boolean;
function  CheckDxx(pst : str16) : boolean;
function  DIRsector(bDev : byte) : byte;
function  DIRtrack(bDev : byte) : byte;
function  DiskFreeDxx(H : integer) : Longint;
function  DiskIDcode(H : integer) : string;
function  DxxExtension(H : integer) : string;
function  Emu2type(Emu : TEmuType) : str16;
function  FindFile(H : integer) : Boolean;
function  FindFreeSector(H : integer) : boolean;
function  FType2txt(by : byte) : string;
function  InfoTrack(H : integer) : byte;
function  Label2Filename(pst, ext : str16; Dir : str80) : str16;
function  MaxSectors(H, Track : Byte) : Byte;
function  MaxTracks(H : integer) : Byte;
function  NoAccess(H : integer) : boolean;
function  SectorIsFree(H, fTrack, fSector : byte) : Boolean;
function  TalkSec(H : integer) : Byte;
function  Translate2AbsPos(H : integer; Track, Sector : Longint) : Longint;
function  Txt2FType(s : Str16) : byte;
function  WildcardMatch(FName, Card : str80) : Boolean;

procedure  AtnHiIEEE;
procedure  AtnLoIEEE;
procedure  AtnTrapIeeeOff;
procedure  AtnTrapIeeeOn;
procedure  DavHi;
procedure  DavLo;
procedure  EoiHi;
procedure  EoiLo;
procedure  InitIEEE;
procedure  LedOff;
procedure  LedOn;
procedure  NdacHi;
procedure  NdacLo;
procedure  NrfdHi;
procedure  NrfdLo;
procedure  ParallelInput;
procedure  ParallelOutput;
procedure  RenHi;
procedure  RenLo;
procedure  ResetHiIEEE;
procedure  ResetLoIEEE;

procedure  ByteOut(H, Value : Byte; DoEOI : Boolean);
procedure  ByteOut2(H, Value : Byte; DoEOI : Boolean);
procedure  ChangeDirectory(H : integer);
procedure  CloseAllHandles(bDevNo, bDrvNo : Integer);
procedure  CloseHandle(var H : Integer);
procedure  CreateSubdirD81(H : integer; Name : str16; fTrack : byte; fNumSectors : word);
procedure  DriveMsg(H, E, Track, Sector : word);
procedure  FindFirstDxx(H : integer);
procedure  FindNextDxx(H : integer);
procedure  GCRdecode(var GCR : TGCRbytes);
procedure  GCRencode(var GCR : TGCRbytes);
procedure  GetDirEntry(H : integer; Var DirEntryDxx : TDirEntryDxx);
procedure  InitDxx(H : integer);
procedure  PutDirEntry(H : integer; Var DirEntryDxx : TDirEntryDxx);
procedure  ReadBlockDxx(H : Integer; ReadData : Word;
                          Var Data; Var DataRead : Word; var pbo : boolean);
procedure  ReadSector(H, T, S : byte);
procedure  ResetSector(H, fTrack, fSector : byte);
procedure  ScratchDxx(H : integer);
procedure  SeekBAMentry(H, bT, bS : byte; var BAMpos : TBAMpos);
procedure  SetSector(H, fTrack, fSector : byte);
procedure  Validate(H : integer);
procedure  WriteBlockDxx(H : Integer; WriteData : Word;
                               Var Data; Var DataWritten : Word);
procedure  WriteSector(H, T, S : byte);



{============================================================================}


implementation


uses CRT, DOS, CBM_unit;


{============================================================================}


function  FindFreeDirSector(H : integer) : boolean;                  forward;
procedure  SubSector(H, fTrack, fSector : byte; pbo : boolean);      forward;
procedure  ValidateDir(H, fTrack, fSector : byte);                   forward;


{============================================================================}


(*
procedure Debug(H : integer);
var
  DxxFile : File;
begin
  with DriveHandles[H] do with DebugR do
   begin
    ThbDrvNo := hbDrvNo;
    ThbDevNo := hbDevNo;
    ThSlot := hSlot;
    ThdSector := hdSector;
    ThdTrack := hdTrack;
    ThSector := hSector;
    ThTrack := hTrack;

    with Device[hbDevNo]^ do
     begin
      tErrMsg := dErrorMsg;

      with Drv[hbDrvNo] do
       begin
        tPCFile := PCfile;
        Tdatablok := dDataBlock;
        TdSector := dSector;
        TdTrack := dTrack;
       end;
     end;
   end;

 for tne := 0 to 1000 do tv := port[$278];

{
  with DriveHandles[H] do
   begin
    Assign(DxxFile, Device[hbDevNo]^.Drv[hbDrvNo].PCfile);
    if not exists(Device[hbDevNo]^.Drv[hbDrvNo].PCfile) then
     writeln(Device[hbDevNo]^.Drv[hbDrvNo].PCfile);
   end;

  Reset(DxxFile, 1);
  Seek(DxxFile, 91392);

  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do BlockRead(DxxFile, tBuf, 10);

  Close(DxxFile);

  if (tBuf[0] <> 18) or (tBuf[1] <> 1) or (tBuf[2] <> 65) then
   writeln('!!!!!!!!!!!!!');
}
end;
*)



{============================================================================}



function AddDirEntry(H : Integer) : boolean;
{---------------------------------------------------------------------------
Adds another entry into the directory
---------------------------------------------------------------------------}
Var
  pbo : boolean;
  OldS, OldT, B, T, S : byte;
  K : integer;
  FName : Str16;
  DirEntryDxx : TDirEntryDxx;

begin
  AddDirEntry := False;

 with DriveHandles[H] do
  begin
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    begin
     B := hFiletype;                            { save wanted type }
     hFileType := $FF;              {#### nog $.. invullen? }

{  Delete old entry first  }
     if FindFile(H) then
      begin
       if    (not hReplace)
          or (B <> hFileType) then              { compare with found type }
        begin
         DiskStatus := -3;                      { File exists }
         DriveMsg(H, 63, 0, 0);
         hFileType := B;
         Exit;
        end;

       ScratchDxx(H);
       if Error <> 0 then exit;
      end;
    end;

   if (hDrvNum and 1 = 1) then hbDrvNo := 0 else hbDrvNo := 1;

   with Device[hbDevNo]^.Drv[hbDrvNo] do
    begin

{  First create dir entry  }
     hFiletype := B;
     FName := hFilename;
     DiskStatus := 0;
     Fillchar(FName[Length(FName)+1], 16-Length(FName), $A0);

     hdTrack := dTrack;
     hdSector := dSector;
     hSlot := 0;

     repeat
      GetDirEntry(H, DirEntryDxx);

      if (DirEntryDxx.Filetype = 0) then         { if erased then... }
       begin
        Fillchar(DirEntryDxx, Sizeof(DirEntryDxx), 0);
        DirEntryDxx.Filetype := hFiletype;

   {  If subdirectory then store parent in bytes for relative data  }
        if (hFiletype = $86) then
         begin
          DirEntryDxx.TSideSec := dTrack;
          DirEntryDxx.SSideSec := dSector;
         end;

        Move(FName[1], DirEntryDxx.FName, 16);

   {  Write it back for the moment  }
        PutDirEntry(H, DirEntryDxx);
        Break;
       end;

   {  Calculate next entry  }
      if hSlot = 7 then
       begin
        OldS := hdSector;
        OldT := hdTrack;
        ReadSector(H, OldT, OldS);
        hdTrack := dDataBlock[0];
        hdSector := dDataBlock[1];
        hSlot := 0;

        if dDataBlock[0] = 0 then

     {  This was last used sector for directory, add another one  }
         begin
          K := AssignHandle(H, 'xxx');
          DriveHandles[K] := DriveHandles[H];

          if not FindFreeDirSector(K) then
           begin
            DiskStatus := -2;                     { Error: Directory is full }
            Exit;
           end;

     {  Put sector in the chain  }
          ReadSector(H, OldT, OldS);
          dDataBlock[0] := DriveHandles[K].hTrack;
          dDataBlock[1] := DriveHandles[K].hSector;
          hdTrack := dDataBlock[0];
          hdSector := dDataBlock[1];
          WriteSector(H, OldT, OldS);
          CloseHandle(K);

     {  Prepare new sector  }
          Fillchar(dDataBlock, Sizeof(dDataBlock), 0);
          for B := 0 to 255 do dDataBlock[B] := 0; { Clear complete sector }
          dDataBlock[1] := 255;                    { Complete sector in use }
          WriteSector(H, hdTrack, hdSector);
         end
          else ReadSector(H, hdTrack, hdSector);
       end
        else Inc(hSlot);
     until False;

     if (DiskStatus <> 0) then exit;               { Can't create dir entry }

     DirEntryDxx.Track_1  := 0;
     DirEntryDxx.Sector_1 := 0;
     DirEntryDxx.Lo_Size  := 0;
     DirEntryDxx.Hi_Size  := 0;
     PutDirEntry(H, DirEntryDxx);

     hBlockSize := 0;
     hTrack     := 0;
     hSector    := 0;
     hOffset    := 0;
     hPosition  := 0;
     HAccess    := fWRITE;
    end;
  end;

  AddDirEntry := True;
end;


function AllocateSector(H : integer) : boolean;
{---------------------------------------------------------------------------
Allocates a sector to the newly made directory entry
---------------------------------------------------------------------------}
var
  DirEntryDxx : TDirEntryDxx;
begin
  if not FindFreeSector(H) then with DriveHandles[H] do
   begin
    Device[hbDevNo]^.Drv[hbDrvNo].DiskStatus := -2;  { Error: Disk is full }
    DriveMsg(H, 72, 0, 0);
    AllocateSector := False;
    Exit;
   end;

  GetDirEntry(H, DirEntryDxx);
  DirEntryDxx.Track_1  := DriveHandles[H].hTrack;
  DirEntryDxx.Sector_1 := DriveHandles[H].hSector;
  DirEntryDxx.Lo_Size  := 1;                  { start with 1 sector }
  DirEntryDxx.Hi_Size  := 0;
  PutDirEntry(H, DirEntryDxx);
  AllocateSector := True;

  with DriveHandles[H] do
   begin
    hBlockSize  := 1;
    hOffset     := 2;          { Skip first two bytes (point to next sector) }
    hPosition   := 0;                           { At begining of file }
    HAccess     := fWRITE;
   end;
end;


function AssignHandle(H : integer; FName : Str16) : Integer;
{---------------------------------------------------------------------------
Creates a H for the given filename
0 is reserved for temporary handling
---------------------------------------------------------------------------}
Var
  I : Integer;
begin
  AssignHandle := -1;                           { Invalid H }

  for I := 1 to coMaxDriveHandles do with DriveHandles[I] do
   if ((hFilename = FName) and (HAccess = fWRITE)) then
    begin
     DriveMsg(H, 60, 0, 0);                     { write file open }
     Exit;
    end;

  for I := 1 to coMaxDriveHandles do
   if DriveHandles[I].hFilename = '' then
    begin
     DriveHandles[I] := DriveHandles[H];

     with DriveHandles[I]do
      begin
       hFilename := FName;
       HAccess   := fREAD;
       hdSector  := DIRsector(hbDevNo);
       hdTrack   := DIRtrack(hbDevNo);
       hBlkPnt   := 0;
       hForceDel := False;
       AssignHandle := I;
      end;

     exit;
    end;

  DriveMsg(H, 78, 0, 0);                        { NO H AVAILABLE }
end;


function BAMsector(H, b : byte) : byte;
begin
  case Device[DriveHandles[H].hbDevNo]^.FStype of
   Emu3040,
   Emu4040,
   Emu1541,
   Emu1571 : BAMsector := 0;
   Emu1581 : BAMsector := b;
   Emu8050,
   Emu8250 : BAMsector := b * 3;
   EmuD16  : BAMsector := 224 + b;
  end;
end;


function BAMtrack(H, b : byte) : byte;
begin
  case Device[DriveHandles[H].hbDevNo]^.FStype of
    Emu3040,
    Emu4040,
    Emu1541,
    Emu1571 : BAMtrack := 18 + (b * 35);
    Emu1581 : BAMtrack := 40;
    Emu8050,
    Emu8250 : BAMtrack := 38;
    EmuD16  : BAMtrack := 18;
   end;
end;


function ByteIn(H : integer; Var EOI : Boolean) : Byte;
{-----------------------------------------------------------------------------
Receive a byte from the external CBM drive
Out: AL: the byte received
-----------------------------------------------------------------------------}
const

{*  load"$",x  *}
  aST1a : array[0..2] of Str23 = (#$28#$F0, '$0', #$3F#$48#$60#$5F);
  coN1a = 3;
  aST1b : array[0..2] of Str23 = (#$29#$F0, '$0', #$3F#$49#$60#$5F);
  coN1b = 3;
  aST1c : array[0..2] of Str23 = (#$2A#$F0, '$0', #$3F#$4A#$60#$5F);
  coN1c = 3;
  aST1d : array[0..2] of Str23 = (#$2B#$F0, '$0', #$3F#$4B#$60#$5F);
  coN1d = 3;


{*  CBM = converting DOS directory to C= format  *}
  aST2 : array[0..2] of Str23 = (#$2A#$FF, 'CBM', #$3F);
  coN2 = 3;


{*  save"test",x  *}
  aST3a : array[0..3] of Str23 = (#$28#$F1, 'TEST',
          #$28#$61#1#8#1#0#7#8'ABCD'#0#0#0, #$3F);
  coN3a = 4;
  aST3b : array[0..3] of Str23 = (#$29#$F1, 'TEST',
          #$29#$61#1#8#1#0#7#8'ABCD'#0#0#0, #$3F);
  coN3b = 4;
  aST3c : array[0..3] of Str23 = (#$2A#$F1, 'TEST',
          #$2A#$61#1#8#1#0#7#8'ABCD'#0#0#0, #$3F);
  coN3c = 4;
  aST3d : array[0..3] of Str23 = (#$2B#$F1, 'TEST',
          #$2B#$61#1#8#1#0#7#8'ABCD'#0#0#0, #$3F);
  coN3d = 4;

  aST4 : array[0..3] of Str23 = (#$28#$F1, 'TEST567890123456',
          #$28#$61#1#8#1#0#7#8'ABCD'#0#0#0, #$3F);
  coN4 = 4;

 { name too long, should be shortened }
  aST5 : array[0..3] of Str23 = (#$28#$F1, 'TEST567890123456789',
          #$28#$61#1#8#1#0#7#8'ABCD'#0#0#0, #$3F);
  coN5 = 4;

Var
  pbo : boolean;
  DByte, B, I, WCoD : Byte;
  st                : Str80;
  ECnt, W1          : Word;
label L2;
begin
  if boTest then
   begin
    B  := coN1c;
    st := Ast1c[bByteInCnt1];

    ByteIn := ord(st[bByteInCnt2]);
    EOI := (bByteInCnt2 = length(st));

    if EOI then
     begin
      inc(bByteInCnt1);
      bByteInCnt2 := 1
     end
    else inc(bByteInCnt2);

    boTestEnd := (bByteInCnt1 = B);
    exit;
   end;

 asm
  cli
 end;

  EOI := False;
  ParallelInput;
  NdacLo;
  NrfdHi;
  AtnTrapIeeeOff;

{ Wait for DAV = (L) }
  ECnt := 40000;

  while not DavFound do
   begin
    dec(ECnt);

    if ECnt = 0 then
     begin
write('RO1 ');
      goto L2;                                { timeout, -> }
     end;
   end;

  NrfdLo;

  repeat
   Dbyte := port[wIeeeLptAddr];
  until Dbyte = port[wIeeeLptAddr];

  EOI := EoiFound;
  NdacHi;

{ Wait for DAV = (H) }
  ECnt := 40000;
  while DavFound do
   begin
    dec(ECnt);

    if ECnt = 0 then
     begin
write('RO2 ');
      goto L2;                     { timeout, -> }
     end;
   end;

asm
sti
end;

  NdacLo;
  ByteIn := not Dbyte;
  AtnTrapIeeeOn;

{####}
{write('  ', Dec2Hex(not Dbyte, 2));}

  exit;

L2:
  Error := $80;
  NdacHi;
  ByteIn := 0;

 asm
  sti
 end;
end;


function CheckDxx(pst : str16) : boolean;
begin
  pst := Upper(pst);
  CheckDxx := (   (pst = 'D64')
               or (pst = 'D16')
               or (pst = 'D71')
               or (pst = 'D80')
               or (pst = 'D81')
               or (pst = 'D82')
               or (pst = 'D30'));
end;


function CreateDiskImage(H : integer; stLabel, stID : str16) : boolean;
const
  Sect40 : array[0..5] of byte = ($28,$FF,$FF,$FF,$FF,$FF);
  Sect29 : array[0..4] of byte = ($1D,$FF,$FF,$FF,$1F);
  Sect27 : array[0..4] of byte = ($1B,$FF,$FF,$FF,$07);
  Sect25 : array[0..4] of byte = ($19,$FF,$FF,$FF,$01);
  Sect23 : array[0..4] of byte = ($17,$FF,$FF,$7F,$00);
  Sect21 : array[0..3] of byte = ($15,$FF,$FF,$1F);
  Sect20 : array[0..3] of byte = ($14,$FF,$FF,$0F);
  Sect19 : array[0..3] of byte = ($13,$FF,$FF,$07);
  Sect18 : array[0..3] of byte = ($12,$FF,$FF,$03);
  Sect17 : array[0..3] of byte = ($11,$FF,$FF,$01);
var
  bDev, bDrv : byte;
  b2, b3 : byte;
  f1 : file;
  fb : file of byte;
  stPCname : str80;
  BAMpos : TBAMpos;
  wNBlocks, w1, w2, w3 : word;
label PL1;
begin
  if Device[bDev]^.FStype = EmuPC then
   begin
    CreateDiskImage := True;
    exit;
   end;

  with DriveHandles[H] do
   begin
    bDev := hbDevNo;
    bDrv := hbDrvNo;
   end;

  if Device[bDev]^.FStype = EmuCBM then
   begin
    CreateDiskImage := True;
    exit;
   end;

  CreateDiskImage := False;
  stPCname := Device[bDev]^.Drv[bDrv].PCfile;

  case Device[bDev]^.FStype of
    Emu1541 : wNBlocks := 683;
    Emu1571 : wNBlocks := 1366;
    Emu1581 : wNBlocks := 3200;
    Emu3040 : wNBlocks := 690;
    Emu4040 : wNBlocks := 683;
    Emu8050 : wNBlocks := 2083;
    Emu8250 : wNBlocks := 4166;
    EmuD16  : wnBlocks := $FE00;                      { 255 * 256 }
   end;

  assign(f1, stPCname);
  {$I-} rewrite(f1, 256); {$I+}
  if (IOresult > 0) then exit;

  with Device[bDev]^.Drv[bDrv] do
   begin
    for w1 := 0 to 255 do dDataBlock[w1] := 0;
    for w1 := 1 to wNBlocks do
     begin
      {$I-} BlockWrite(f1, dDataBlock, 1); {$I+}
      if (IOresult > 0) then exit;
     end;
   end;

  close(f1);

  assign(fb, stPCname);
  {$I-} reset(fb); {$I+}
  if (IOresult > 0) then exit;

  with Device[bDev]^.Drv[bDrv] do
  case Device[bDev]^.FStype of
    Emu1541, Emu4040 : begin                    { D64 }
PL1:  dDataBlock[0] := 18;
      dDataBlock[1] := 1;
      dDataBlock[2] := ord('A');

      for w3 := 0 to 3 do
       begin
        for w2 :=  1 to 17 do dDataBlock[w2*4 + w3] := Sect21[w3];
        for w2 := 18 to 24 do dDataBlock[w2*4 + w3] := Sect19[w3];
        for w2 := 25 to 30 do dDataBlock[w2*4 + w3] := Sect18[w3];
        for w2 := 31 to 35 do dDataBlock[w2*4 + w3] := Sect17[w3];
       end;

      dDataBlock[72] := 17;                     { Track 18 only 17 free }
      dDataBlock[73] := $FC;                    { Sector 0 and 1 used }

  {* Label *}
      for w3 := 1 to length(stLabel) do
       dDataBlock[143 + w3] := ord(stLabel[w3]);
      for w3 := length(stLabel)+1 to 18 do      { add shifted space + 2 }
       dDataBlock[143 + w3] := 160;

  {* ID *}
      dDataBlock[$A2] := ord(stID[1]);
      dDataBlock[$A3] := ord(stID[2]);
      dDataBlock[$A5] := ord(stID[4]);
      dDataBlock[$A6] := ord(stID[5]);
      if (stID[3] = ' ') then dDataBlock[$A4] := 160 else
       dDataBlock[$A4] := ord(stID[3]);

      for w3 := $A7 to $AA do dDataBlock[w3] := 160;

      seek(fb, Translate2AbsPos(H, 18, 0));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);

  {* Track 18 sector 1, first directory sector *}
      b2 := 0;                                  { first and only sector }
      b3 := $FF;                                { all bytes used }
      write(fb, b2);
      write(fb, b3);
      for w3 := 0 to 254 do write(fb, b2);      { clear rest of sector }
     end;

    Emu1571 : begin                                   { D71 }

  {* Create BAM for track 36 - 70 *}
      for w3 := 0 to 255 do dDataBlock[w3] := 0;

      for w3 := 1 to 3 do
       begin
        for w2 :=  1 to 17 do dDataBlock[w2*3 + w3 - 4] := Sect21[w3];

    {* Start at 19: track 53 may not be used *}
        for w2 := 19 to 24 do dDataBlock[w2*3 + w3 - 4] := Sect19[w3];
        for w2 := 25 to 30 do dDataBlock[w2*3 + w3 - 4] := Sect18[w3];
        for w2 := 31 to 35 do dDataBlock[w2*3 + w3 - 4] := Sect17[w3];
       end;

      seek(fb, Translate2AbsPos(H, 53, 0));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);

  {* First BAM sector *}
      for w1 := 0 to 255 do dDataBlock[w1] := 0;;
      dDataBlock[3] := 1;                        { double sided }

  {* Free-sector bytes in first part of BAM  *}
      for w2 :=  1 to 17 do dDataBlock[$DD + w2] := Sect21[0];

    {* Start at 19: track 53 may not be used *}
      for w2 := 19 to 24 do dDataBlock[$DD + w2] := Sect19[0];
      for w2 := 25 to 30 do dDataBlock[$DD + w2] := Sect18[0];
      for w2 := 31 to 35 do dDataBlock[$DD + w2] := Sect17[0];

    {* Use D64 routines further *}
      goto PL1;
     end;

    Emu1581 : begin                                   { D81 }

  {  Track 40, sector 0  }
      dDataBlock[0] := 40;
      dDataBlock[1] := 3;
      dDataBlock[2] := ord('D');
      dDataBlock[3] := 0;

  {* Label *}
      for w3 := 1 to length(stLabel) do
       dDataBlock[3 + w3] := ord(stLabel[w3]);
      for w3 := length(stLabel)+1 to 18 do      { add shifted space + 2 }
       dDataBlock[3 + w3] := 160;

  {* ID *}
      dDataBlock[$16] := ord(stID[1]);
      dDataBlock[$17] := ord(stID[2]);
      dDataBlock[$19] := ord(stID[4]);
      dDataBlock[$1A] := ord(stID[5]);
      if (stID[3] = ' ') then dDataBlock[$18] := 160 else
       dDataBlock[$18] := ord(stID[3]);

      for w3 := $1B to $1C do dDataBlock[w3] := 160;

      seek(fb, Translate2AbsPos(H, 40, 0));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);

  {  Track 40, sector 1 = first BAM-sector  }
      for w3 := 0 to 255 do dDataBlock[w3] := 0;

      dDataBlock[0] := 40;
      dDataBlock[1] := 2;
      dDataBlock[2] := ord('D');
      dDataBlock[3] := $BB;
      dDataBlock[4] := ord(stID[4]);
      dDataBlock[5] := ord(stID[5]);
      dDataBlock[6] := $C0;                      { I/O byte }

      for w3 := 0 to 39 do
       for w2 := 0 to 5 do dDataBlock[w3*6 + w2 + 16] := Sect40[w2];

      SeekBAMentry(H, 40, 0, BAMpos);
      w1 := BAMpos.bByt - 1;
      dDataBlock[w1] := $24;                     { Track40 Sector 0..3 used }
      dDataBlock[w1 + 1] := $F0;
      for w2 := 0 to 255 do write(fb, dDataBlock[w2]);

  {  Track 40, sector 2 = second BAM-sector  }
      dDataBlock[0] := 0;
      dDataBlock[1] := $FF;
      dDataBlock[w1] := Sect40[0];
      dDataBlock[w1 + 1] := Sect40[1];
      for w2 := 0 to 255 do write(fb, dDataBlock[w2]);

  {* Track 40 sector 3, first directory sector *}
      b2 := 0;                                  { first and only sector }
      b3 := $FF;                                { all bytes used }
      write(fb, b2);
      write(fb, b3);
     end;

    Emu8050 : begin                                   { D80 }

  {* Directory header *}
   { Track 39, sector 0 }
      dDataBlock[0] := 38;                       { pointer to first }
      dDataBlock[1] := 0;                        { BAM block }
      dDataBlock[2] := ord('C');
      dDataBlock[3] := 0;

   { Label }
      for w3 := 1 to length(stLabel) do
       dDataBlock[5 + w3] := ord(stLabel[w3]);
      for w3 := length(stLabel)+1 to 18 do      { add shifted space + 2 }
       dDataBlock[5 + w3] := 160;

   { ID }
      dDataBlock[24] := ord(stID[1]);
      dDataBlock[25] := ord(stID[2]);
      dDataBlock[27] := ord(stID[4]);
      dDataBlock[28] := ord(stID[5]);
      if (stID[3] = ' ') then dDataBlock[26] := 160 else
       dDataBlock[26] := ord(stID[3]);

      for w3 := 29 to 32 do dDataBlock[w3] := 160;

      seek(fb, Translate2AbsPos(H, 39, 0));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);

  {* Track 39 sector 1, first directory sector *}
      b2 := 0;                                  { first and only sector }
      b3 := $FF;                                { all bytes used }
      write(fb, b2);
      write(fb, b3);
      for w1 := 2 to 255 do write(fb, b2);      { clear all entries }

  {  Track 38, sector 0 = first BAM-sector  }
      for w3 := 0 to 255 do dDataBlock[w3] := 0;

      dDataBlock[0] := 38;
      dDataBlock[1] := 3;
      dDataBlock[2] := ord('C');
      dDataBlock[3] := 0;
      dDataBlock[4] := 1;
      dDataBlock[5] := 51;

      for w3 := 0 to 4 do
       begin
        for w2 :=  1 to 39 do dDataBlock[w2*5 + w3 + 1] := Sect29[w3];
        for w2 := 40 to 50 do dDataBlock[w2*5 + w3 + 1] := Sect27[w3];
       end;

      SeekBAMentry(H, 38, 0, BAMpos);
      w1 := BAMpos.bByt - 1;
      dDataBlock[w1] := 27;                      { track 38 sector 0, 3 used }
      dDataBlock[w1 + 1] := $F6;

      SeekBAMentry(H, 39, 0, BAMpos);
      w1 := BAMpos.bByt - 1;
      dDataBlock[w1] := 27;                      { track 39 sector 0, 1 used }
      dDataBlock[w1 + 1] := $FC;

      seek(fb, Translate2AbsPos(H, 38, 0));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);

  {  Track 38, sector 3 = second BAM-sector  }
      for w3 := 0 to 255 do dDataBlock[w3] := 0;

      dDataBlock[0] := 39;                       { pointer to first }
      dDataBlock[1] := 1;                        { directory block }
      dDataBlock[2] := ord('C');
      dDataBlock[3] := 0;
      dDataBlock[4] := 51;
      dDataBlock[5] := 78;

      for w3 := 0 to 4 do
       begin
        for w2 := 51 to 53 do dDataBlock[w2*5 + w3 - 249] := Sect27[w3];
        for w2 := 54 to 64 do dDataBlock[w2*5 + w3 - 249] := Sect25[w3];
        for w2 := 65 to 77 do dDataBlock[w2*5 + w3 - 249] := Sect23[w3];
       end;

      seek(fb, Translate2AbsPos(H, 38, 3));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);
     end;

    Emu8250 : begin                                   { D82 }

  {* Directory header *}
   { Track 39, sector 0  }
      dDataBlock[0] := 38;                       { pointer to first }
      dDataBlock[1] := 0;                        { BAM block }
      dDataBlock[2] := ord('C');
      dDataBlock[3] := 0;

   { Label }
      for w3 := 1 to length(stLabel) do
       dDataBlock[5 + w3] := ord(stLabel[w3]);
      for w3 := length(stLabel)+1 to 18 do      { add shifted space + 2 }
       dDataBlock[5 + w3] := 160;

   { ID }
      dDataBlock[24] := ord(stID[1]);
      dDataBlock[25] := ord(stID[2]);
      dDataBlock[27] := ord(stID[4]);
      dDataBlock[28] := ord(stID[5]);
      if (stID[3] = ' ') then dDataBlock[26] := 160 else
       dDataBlock[26] := ord(stID[3]);

      for w3 := 29 to 32 do dDataBlock[w3] := 160;

      seek(fb, Translate2AbsPos(H, 39, 0));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);

  {* Track 39 sector 1, first directory sector *}
      b2 := 0;                                  { first and only sector }
      b3 := $FF;                                { all bytes used }
      write(fb, b2);
      write(fb, b3);
      for w1 := 2 to 255 do write(fb, b2);      { clear all entries }

  {* Track 38, sector 0 = first BAM-sector *}
      for w3 := 0 to 255 do dDataBlock[w3] := 0;

      dDataBlock[0] := 38;
      dDataBlock[1] := 3;
      dDataBlock[2] := ord('C');
      dDataBlock[3] := 0;
      dDataBlock[4] := 1;                       { BAM sector covers track 1 }
      dDataBlock[5] := 51;                      {  to track 51 }

      for w3 := 0 to 4 do
       begin
        for w2 :=  1 to 39 do dDataBlock[w2*5 + w3 + 1] := Sect29[w3];
        for w2 := 40 to 50 do dDataBlock[w2*5 + w3 + 1] := Sect27[w3];
       end;

      SeekBAMentry(H, 38, 0, BAMpos);
      w1 := BAMpos.bByt - 1;
      dDataBlock[w1] := 25;                      { track 38 }
      dDataBlock[w1 + 1] := $B6;                 { sector 0, 3, 6 used }
      dDataBlock[w1 + 2] := $FD;                 { sector 9 used }

      SeekBAMentry(H, 39, 0, BAMpos);
      w1 := BAMpos.bByt - 1;
      dDataBlock[w1] := 27;                      { track 39 sector 0, 1 used }
      dDataBlock[w1 + 1] := $FC;

      seek(fb, Translate2AbsPos(H, 38, 0));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);

  {  Track 38, sector 3 = second BAM-sector  }
      for w3 := 0 to 255 do dDataBlock[w3] := 0;

      dDataBlock[0] := 38;
      dDataBlock[1] := 6;
      dDataBlock[2] := ord('C');
      dDataBlock[3] := 0;
      dDataBlock[4] := 51;
      dDataBlock[5] := 101;

      for w3 := 0 to 4 do
       begin
        for w2 := 51 to  53 do dDataBlock[w2*5 + w3 - 249] := Sect27[w3];
        for w2 := 54 to  64 do dDataBlock[w2*5 + w3 - 249] := Sect25[w3];
        for w2 := 65 to  77 do dDataBlock[w2*5 + w3 - 249] := Sect23[w3];
        for w2 := 78 to 100 do dDataBlock[w2*5 + w3 - 249] := Sect29[w3];
       end;

      seek(fb, Translate2AbsPos(H, 38, 3));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);

  {  Track 38, sector 6 = second BAM-sector  }
      for w3 := 0 to 255 do dDataBlock[w3] := 0;

      dDataBlock[0] := 38;
      dDataBlock[1] := 9;
      dDataBlock[2] := ord('C');
      dDataBlock[3] := 0;
      dDataBlock[4] := 101;
      dDataBlock[5] := 151;

      for w3 := 0 to 4 do
       begin
        for w2 := 101 to 116 do dDataBlock[w2*5 + w3 - 499] := Sect29[w3];
        for w2 := 117 to 130 do dDataBlock[w2*5 + w3 - 499] := Sect27[w3];
        for w2 := 131 to 141 do dDataBlock[w2*5 + w3 - 499] := Sect25[w3];
        for w2 := 142 to 150 do dDataBlock[w2*5 + w3 - 499] := Sect23[w3];
       end;

      seek(fb, Translate2AbsPos(H, 38, 6));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);

  {  Track 38, sector 9 = third BAM-sector  }
      for w3 := 0 to 255 do dDataBlock[w3] := 0;

      dDataBlock[0] := 39;                       { pointer to first }
      dDataBlock[1] := 1;                        { directory block }
      dDataBlock[2] := ord('C');
      dDataBlock[3] := 0;
      dDataBlock[4] := 151;
      dDataBlock[5] := 155;

      for w3 := 0 to 4 do
        for w2 := 151 to 154 do dDataBlock[w2*5 + w3 - 749] := Sect23[w3];

      seek(fb, Translate2AbsPos(H, 38, 9));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);
     end;

    EmuD16 : begin                                   { D16 }
      dDataBlock[0] := 18;
      dDataBlock[1] := 1;
      dDataBlock[2] := ord('R');

  {* Label *}
      for w3 := 1 to length(stLabel) do
       dDataBlock[143 + w3] := ord(stLabel[w3]);
      for w3 := length(stLabel)+1 to 18 do      { add shifted space + 2 }
       dDataBlock[143 + w3] := 160;

  {* ID *}
      dDataBlock[$A2] := ord(stID[1]);
      dDataBlock[$A3] := ord(stID[2]);
      dDataBlock[$A5] := ord(stID[4]);
      dDataBlock[$A6] := ord(stID[5]);
      if (stID[3] = ' ') then dDataBlock[$A4] := 160 else
       dDataBlock[$A4] := ord(stID[3]);

      for w3 := $A7 to $AA do dDataBlock[w3] := 160;

      seek(fb, Translate2AbsPos(H, 18, 0));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);

  {* Track 18 sector 1, first directory sector *}
      b2 := 0;                                  { first and only sector }
      b3 := $FF;                                { all bytes used }
      write(fb, b2);
      write(fb, b3);
      for w1 := 2 to 255 do write(fb, b2);      { clear all entries }


  {  Track 18, sector 224 = first BAM-sector  }
      seek(fb, Translate2AbsPos(H, 18, 224));
      b3 := $FF;
      for w1 := 0 to 8159 do write(fb, b3);     { Track 1 .. 255 }

  {  Correct for use of directory  }
      SeekBAMentry(H, 18, 0, BAMpos);
      w1 := BAMpos.Sector;
      w2 := BAMpos.bByt;
      seek(fb, Translate2AbsPos(H, 18, w1) + w2);
      b3 := $FC;
      write(fb, b3);

(*
  {  Correct for use of BAM  }
      SeekBAMentry(H, 18, 224, BAMpos);
      w1 := BAMpos.Sector;
      w2 := BAMpos.bByt;
      seek(fb, Translate2AbsPos(H, 18, w1) + w2);
      b2 := 0;
      for b3 := 1 to 32 do write(fb, b2);
*)
     end;
    Emu3040 : begin                                   { D30 }
      dDataBlock[0] := 18;
      dDataBlock[1] := 1;
      dDataBlock[2] := 1;

      for w3 := 0 to 3 do
       begin
        for w2 :=  1 to 17 do dDataBlock[w2*4 + w3] := Sect21[w3];
        for w2 := 18 to 24 do dDataBlock[w2*4 + w3] := Sect20[w3];
        for w2 := 25 to 30 do dDataBlock[w2*4 + w3] := Sect18[w3];
        for w2 := 31 to 35 do dDataBlock[w2*4 + w3] := Sect17[w3];
       end;

      SeekBAMentry(H, 18, 0, BAMpos);
      w1 := BAMpos.bByt - 1;
      dDataBlock[w1] := 18;                      { Track 18 only 18 free }
      dDataBlock[w1 + 1] := $FC;                 { Sector 0 and 1 used }

  {* Label *}
      for w3 := 1 to length(stLabel) do
       dDataBlock[143 + w3] := ord(stLabel[w3]);
      for w3 := length(stLabel)+1 to 18 do      { add shifted space + 2 }
       dDataBlock[143 + w3] := 160;

  {* ID *}
      dDataBlock[$A2] := ord(stID[4]);
      dDataBlock[$A3] := ord(stID[5]);

      for w3 := $A4 to $AA do dDataBlock[w3] := 160;

      seek(fb, Translate2AbsPos(H, 18, 0));
      for w1 := 0 to 255 do write(fb, dDataBlock[w1]);

  {* Track 18 sector 1, first directory sector *}
      b2 := 0;                                  { first and only sector }
      b3 := $FF;                                { all bytes used }
      write(fb, b2);
      write(fb, b3);
     end;
   end;

  close(fb);
  CreateDiskImage := True;
end;


function DIRsector(bDev : byte) : byte;
{---------------------------------------------------------------------------
Outputs the first sector of the root directory
---------------------------------------------------------------------------}
begin
  if Device[bDev]^.FStype = Emu1581 then DIRsector := 3
   else DIRsector := 1;
end;


function DIRtrack(bDev : byte) : byte;
{---------------------------------------------------------------------------
Outputs the first track of the root directory
---------------------------------------------------------------------------}
begin
  case Device[bDev]^.FStype of
    Emu1541,
    Emu1571,
    Emu3040,
    Emu4040,
    EmuD16  : DIRtrack := 18;
    Emu1581 : DIRtrack := 40;
    Emu8050,
    Emu8250 : DIRtrack := 39;
    EmuCBM,
    EmuPC   : DIRtrack := 0;
   end;
end;


function DiskFreeDxx(H : integer) : Longint;
{----------------------------------------------------------------------------
Calculates number of free blocks on disk.
in:     -
out:    number of free blocks
----------------------------------------------------------------------------}
Var
  bDev, bDrv : byte;
  B, I, J, K : Byte;
  Free : Longint;
begin
  Free := 0;

  with DriveHandles[H] do
   begin
    bDev := hbDevNo;
    bDrv := hbDrvNo;
   end;

  with DriveHandles[H] do
   with Device[bDev]^.Drv[bDrv] do
  case Device[bDev]^.FStype of
    Emu3040,
    Emu4040,
    Emu1541 : begin
      ReadSector(H, BAMtrack(H, 0), BAMsector(H, 0));
      for I :=  1 to 17 do Free := Free + dDataBlock[I * 4];
      for I := 19 to 35 do Free := Free + dDataBlock[I * 4];
     end;
    Emu1571 : begin
      ReadSector(H, BAMtrack(H, 0), BAMsector(H, 0));
      for I :=  1 to 17 do Free := Free + dDataBlock[I * 4];
      for I := 19 to 35 do Free := Free + dDataBlock[I * 4];
      for I :=  1 to 17 do Free := Free + dDataBlock[$DD + I];
      for I := 19 to 35 do Free := Free + dDataBlock[$DD + I];
     end;

{###  klopt dit wel ??? }
    Emu1581 : begin
      ReadSector(H, BAMtrack(H, 0), BAMsector(H, 0));
      for I :=  0 to 38 do Free := Free + dDataBlock[16 + (I * 6)];
      ReadSector(H, BAMtrack(H, 1), BAMsector(H, 1));
      for I :=  0 to 38 do Free := Free + dDataBlock[16 + (I * 6)];
     end;

    Emu8050 : begin
      ReadSector(H, BAMtrack(H, 0), BAMsector(H, 0));
      for I :=  1 to 38 do Free := Free + dDataBlock[1 + I * 5];
      for I := 40 to 50 do Free := Free + dDataBlock[1 + I * 5];
      ReadSector(H, BAMtrack(H, 1), BAMsector(H, 1));
      for I :=  1 to 27 do Free := Free + dDataBlock[1 + I * 5];
     end;
    Emu8250 : begin
      ReadSector(H, BAMtrack(H, 0), BAMsector(H, 0));
      for I :=  1 to 38 do
       Free := Free + dDataBlock[1 + I * 5];
      for I := 40 to 50 do
       Free := Free + dDataBlock[1 + I * 5];
      ReadSector(H, BAMtrack(H, 1), BAMsector(H, 1));
      for I :=  1 to 50 do
       Free := Free + dDataBlock[1 + I * 5];
      ReadSector(H, BAMtrack(H, 2), BAMsector(H, 2));
      for I :=  1 to 50 do
       Free := Free + dDataBlock[1 + I * 5];
      ReadSector(H, BAMtrack(H, 3), BAMsector(H, 3));
      for I :=  1 to  4 do
       Free := Free + dDataBlock[1 + I * 5];
     end;
    EmuD16 : begin
      for J := 0 to 31 do
       begin
        ReadSector(H, BAMtrack(H, J), BAMsector(H, J));
        for I := 0 to 255 do
         begin
          B := dDataBlock[I];
          for K := 0 to 7 do
           if (B and Power2[K] > 0) then inc(Free);
         end;
       end;
     end;
    EmuCBM,
    EmuPC : Free := DiskFree(Device[bDev]^.Drv[bDrv].DosDrive) div 256;
   end;

  DiskFreeDxx := Free;
end;


function DiskIDcode(H : integer) : string;
begin
  case Device[DriveHandles[H].hbDevNo]^.FStype of
   Emu3040 : DiskIDcode := '  ';
   Emu4040,
   Emu1541,
   Emu1571 : DiskIDcode := '2A';
   Emu1581 : DiskIDcode := '2D';
   Emu8050,
   Emu8250 : DiskIDcode := '2C';
   EmuCBM  : DiskIDcode := '2P';
   EmuD16  : DiskIDcode := '2R';
   EmuPC   : DiskIDcode := 'PC';
  end;
end;


function DxxExtension(H : integer) : string;
begin
  case Device[DriveHandles[H].hbDevNo]^.FStype of
    Emu1541,
    Emu4040 : DxxExtension := '.D64';
    Emu1571 : DxxExtension := '.D71';
    Emu1581 : DxxExtension := '.D81';
    Emu3040 : DxxExtension := '.D30';
    Emu8050 : DxxExtension := '.D80';
    Emu8250 : DxxExtension := '.D82';
    EmuD16  : DxxExtension := '.D16';
    EmuP00  : DxxExtension := '.P00';
    EmuT64  : DxxExtension := '.T64';
   end;
end;


function Emu2type(Emu : TEmuType) : str16;
begin
  case Emu of
    Emu1541 : Emu2type := '1541';
    Emu1571 : Emu2type := '1571';
    Emu1581 : Emu2type := '1581';
    Emu3040 : Emu2type := '3040';
    Emu4040 : Emu2type := '4040';
    Emu8050 : Emu2type := '8050';
    Emu8250 : Emu2type := '8250';
    EmuCBM  : Emu2type := 'CBM-DOS';
    EmuCOM  : Emu2type := 'COM';
    EmuD16  : Emu2type := 'D16';
    EmuLPT  : Emu2type := 'LPT';
    EmuP00  : Emu2type := 'P00-files';
    EmuPC   : Emu2type := 'DOS';
    EmuT64  : Emu2type := 'Tape';
   end;
end;


function FindFile(H : integer) : Boolean;
{---------------------------------------------------------------------------
Finds file in directory
in:     H
out:    yes/no
---------------------------------------------------------------------------}
var
  b, bDev, bDrv, bFileType : byte;
  fb : file of byte;
  k : integer;
  DirInfo  : SearchRec;
  s3 : str16;
  pst, s1, s2 : str80;

 procedure SubFindCBM(bDrvN : byte);
 var ft : text;
 begin
   if not Device[bDev]^.drv[bDrvN].InUse then exit;

   with DriveHandles[H] do
    begin
     ChDir(Device[hbDevNo]^.Drv[0].sDirPath);

     if not exists(stCBMdir) then
      begin
       DriveMsg(H, 62, 0, 1);
       ChDir(OrgDir);
       exit;
      end;

     assign(ft, stCBMdir);
     reset(ft);

     while not EOF(ft) do
      begin
       readln(ft, pst);
       s1 := short(copy(pst, 1, 40));   { DOS name }
       delete(pst, 1, 41);
       s2 := short(copy(pst, 1, 16));   { CBM name }
       delete(pst, 1, 17);
       s3 := copy(pst, 1, 4);           { CBM type }

       if     WildcardMatch(s2, hFilename)
          and (hFiletype and bFileType and $9F = Txt2Ftype(s3)) then
        begin
         Device[hbDevNo]^.Drv[0].DiskStatus := 100;
         Break
        end;
      end;

     ChDir(OrgDir);
     close(ft);
    end;
 end;

 procedure SubFindPC(bDrvN : byte);
 begin
   if not Device[bDev]^.drv[bDrvN].InUse then exit;

   with DriveHandles[H] do
    begin
     ChDir(Device[hbDevNo]^.Drv[0].sDirPath);
     FindFirst('*.*', $37, DirInfo);

     while (DosError = 0) do
      if (DirInfo.Attr and $10 = $10) then
       begin
        if (DirInfo.Name <> '..') and (DirInfo.Name <> '.') then
         begin                          { skip . and .. }
          hFiletype := $86;
          hBlockSize := 1;

          if     WildcardMatch(DirInfo.Name, hFilename)
             and (hFiletype and bFileType = hFileType) then
           begin
            Device[hbDevNo]^.Drv[0].DiskStatus := 100;
            Break
           end;
         end;

        FindNext(DirInfo);
       end
      else
       begin
        k := pos('.', DirInfo.Name);

        if k = 0 then pst := ''
         else pst := copy(DirInfo.Name, k, length(DirInfo.Name)-k+1);

        if pst = '.DEL' then hFiletype := $80 else
        if pst = '.SEQ' then hFiletype := $81 else
        if pst = '.PRG' then hFiletype := $82 else
        if pst = '.USR' then hFiletype := $83 else
        if pst = '.REL' then hFiletype := $84
         else hFiletype := $86;

        if (DirInfo.Attr and 1 > 0) then hFiletype := hFiletype or $40;

        assign(fb, DirInfo.Name);
        {$I-}
        reset(fb);
        {$I+}

        hBlockSize := Filesize(fb) div 254 + 1;

        if hFiletype = $84 then
         begin
          read(fb, b);
          hRecLen := b;
         end
        else hRecLen := 0;

        close(fb);
        if hFiletype and $DF < $86 then DirInfo.Name[0] := chr(k - 1);

        if     WildcardMatch(DirInfo.Name, hFilename)
           and (hFiletype and bFileType and $9F = hFileType and $9F) then
         begin
          Device[hbDevNo]^.Drv[0].DiskStatus := 100;
          Break
         end;

        FindNext(DirInfo);
       end;
    end;
 end;

 procedure SubFind(bDrvN : byte);
 begin
   if not Device[bDev]^.drv[bDrvN].InUse then exit;
   DriveHandles[H].hbDrvNo := bDrvN;
   DriveHandles[K] := DriveHandles[H];
{%%%%   InitDxx(K);}
   FindFirstDxx(K);

   with Device[bDev]^.Drv[bDrvN] do
   repeat
    FindNextDxx(K);

   with DriveHandles[K] do
    if (    (Device[DriveHandles[H].hbDevNo]^.Drv[bDrvN].DiskStatus = 0)
        and WildcardMatch(hFilename, DriveHandles[H].hFilename)
        and (   (hFiletype and bFileType = hFileType and $9F)
             or (hFiletype = $FF)
            )
       ) then
     begin
      DriveHandles[H].hFilename := DriveHandles[K].hFilename;
      DiskStatus := 100;
     end;
   until DiskStatus <> 0;
 end;

begin
  with DriveHandles[H] do
   begin
    bDev := hbDevNo;
    bDrv := hbDrvNo;
    bFileType := hFileType and $9F;
    FindFile := false;
    Device[hbDevNo]^.Drv[0].DiskStatus := 0;
    Device[hbDevNo]^.Drv[1].DiskStatus := 0;
    if hFilename = '' then exit;

    if Device[bDev]^.FStype = EmuPC then
     begin
      if (hDrvNum and 1 = 1) then SubFindPC(0);

      if (    (hDrvNum and 2 = 2)
          and (Device[hbDevNo]^.Drv[0].DiskStatus <> 100)) then
       SubFindPC(1);

      if    (Device[hbDevNo]^.Drv[0].DiskStatus = 100)
         or (Device[hbDevNo]^.Drv[1].DiskStatus = 100) then
       begin
        hFilename := DirInfo.Name;
        hOffset := 2;                         { Skip first two bytes }
        hPosition := 0;                       { Begin of file }
        FindFile := True;
        Device[hbDevNo]^.Drv[0].DiskStatus := 0;
        Device[hbDevNo]^.Drv[1].DiskStatus := 0;

{####  filetype }

       end;

      ChDir(OrgDir);
     end
      else

    if Device[bDev]^.FStype = EmuCBM then
     begin
      if (hDrvNum and 1 = 1) then SubFindCBM(0);

      if (    (hDrvNum and 2 = 2)
          and (Device[hbDevNo]^.Drv[0].DiskStatus <> 100)) then
       SubFindPC(1);

      if    (Device[hbDevNo]^.Drv[0].DiskStatus = 100)
         or (Device[hbDevNo]^.Drv[1].DiskStatus = 100) then
       begin
        hFilename := s3;
        hOffset := 2;                         { Skip first two bytes }
        hPosition := 0;                       { Begin of file }
        FindFile := True;
        Device[hbDevNo]^.Drv[0].DiskStatus := 0;
        Device[hbDevNo]^.Drv[1].DiskStatus := 0;

{####  filetype }

       end;

      ChDir(OrgDir);
     end
    else
     begin
      K := AssignHandle(H, 'xxx');

      if (hDrvNum and 1 = 1) then SubFind(0);

      if (    (hDrvNum and 2 = 2)
          and (Device[hbDevNo]^.Drv[0].DiskStatus <> 100)) then
       SubFind(1);

      if    (Device[hbDevNo]^.Drv[0].DiskStatus = 100)
         or (Device[hbDevNo]^.Drv[1].DiskStatus = 100) then
       begin
        DriveHandles[H] := DriveHandles[K];
        with DriveHandles[H] do
         begin
          hOffset    := 2;        { Skip first two bytes (point to next sector) }
          hPosition  := 0;                           { Begin of file }
          FindFile := True;
          Device[hbDevNo]^.Drv[0].DiskStatus := 0;
          Device[hbDevNo]^.Drv[1].DiskStatus := 0;
         end;
       end;

      CloseHandle(K);
     end;
   end;
end;


function FindFreeDirSector(H : integer) : boolean;
{----------------------------------------------------------------------------
Looks for an unused sector on directory track. D16 allows other tracks to be
used as well. Don't know what to do with D81 yet if track is full.

with the exception of D81 and D16 all other drives use an interleave of 3
sectors.

in:     track/sectro of (sub-) directory
out:    found / not found
        track, sector
----------------------------------------------------------------------------}
Var
  bDrv, bDev : byte;
  Found : boolean;
begin
  with DriveHandles[H] do
   begin
    bDev := hbDevNo;
    bDrv := hbDrvNo;
    hTrack := Device[bDev]^.Drv[bDrv].dTrack;
    hSector := Device[bDev]^.Drv[bDrv].dSector;

    case Device[hbDevNo]^.FStype of
      Emu4040,
      Emu1571,
      Emu1541 : begin
        repeat
         if (hSector > 18) then
          case hSector of
            19 : hSector := 2;
            20 : hSector := 3;
           end;

         Found := SectorIsFree(H, hTrack, hSector);
         if not Found then inc(hSector, 3);
        until (Found or (hSector = 21));
       end;
      Emu1581 : begin
        repeat
         Found := SectorIsFree(H, hTrack, hSector);
         inc(hSector);
        until (Found or (hSector = 40));
        dec(hSector);
       end;
      Emu8050,
      Emu8250 : begin
        repeat
         Found := SectorIsFree(H, hTrack, hSector);
         if not Found then
          begin
           inc(hSector, 3);
           if (hSector > 28) then dec(hSector, 29);
          end;
        until (Found or (hSector = 0));
       end;
      EmuD16  : begin
        repeat
         repeat
          Found := SectorIsFree(H, hTrack, hSector);
          inc(hSector);
         until (Found or (hSector = 0));
         inc(hTrack);
        until (Found or (hTrack = 0));
        if Found then dec(hSector)
         else FindFreeSector(H);
       end;
      Emu3040 : begin
        repeat
         if (hSector > 19) then
          case hSector of
            20 : hSector := 3;
            22 : hSector := 2;
           end;

         Found := SectorIsFree(H, hTrack, hSector);
         if not Found then inc(hSector, 3);
        until (Found or (hSector = 21));
       end;
     end;

    if Found then SetSector(H, hTrack, hSector);
   end;

  FindFreeDirSector := Found;
end;


function FindFreeSector(H : integer) : boolean;
{----------------------------------------------------------------------------
Looks for an unused sector on disk (except directory track)

with the exception of D81 and D16 all other drives use an interleave of 10
or 11 sectors.

in:     -
out:    found / not found
        sector
----------------------------------------------------------------------------}
Var
  bDev, bDrv : byte;
  Found : boolean;
  Tpointer : integer;
begin
  with DriveHandles[H] do
   begin
    bDev := hbDevNo;
    bDrv := hbDrvNo;
    hTrack := DIRtrack(bDev);
    hSector := DIRsector(bDev);
    Tpointer := 1;

    case Device[bDev]^.FStype of
      Emu1541, Emu4040 : begin
        inc(hTrack);

        repeat
         hSector := 0;

         repeat
          Found := SectorIsFree(H, hTrack, hSector);

          if not Found then
           begin
            if (hTrack in [24..30]) then inc(hSector, 11)
             else inc(hSector, 10);

            if (hSector > MaxSectors(H, hTrack)-1) then
              dec(hSector, MaxSectors(H, hTrack));
           end;
         until (Found or (hSector = 0));

         if (Tpointer > 0) then Tpointer := (-1 * (Tpointer + 1))
          else Tpointer := (-1 * (Tpointer - 1));

         if not Found then hTrack := hTrack + Tpointer;
        until (Found or (hTrack = 36));
       end;
      Emu1571 : begin
        inc(hTrack);

        repeat
         hSector := 0;

         repeat
          Found := SectorIsFree(H, hTrack, hSector);

          if not Found then
           begin
            if (hTrack in [24..30]) then inc(hSector, 11)
              else inc(hSector, 10);
            if (hSector > MaxSectors(H, hTrack)-1) then
              dec(hSector, MaxSectors(H, hTrack));
           end;
         until (Found or (hSector = 0));

         if (Tpointer > 0) then Tpointer := (-1 * (Tpointer + 1))
          else Tpointer := (-1 * (Tpointer - 1));

         if not Found then hTrack := hTrack + Tpointer;
        until (Found or (hTrack = 36));

        if not Found then
         begin
          hTrack := 54;
          Tpointer := 1;

          repeat
           hSector := 0;

           repeat
            Found := SectorIsFree(H, hTrack, hSector);

            if not Found then
             begin
              if (hTrack in [59..65]) then inc(hSector, 11)
                else inc(hSector, 10);
              if (hSector > MaxSectors(H, hTrack)-1) then
                dec(hSector, MaxSectors(H, hTrack));
             end;
           until (Found or (hSector = 0));

           if (Tpointer > 0) then Tpointer := (-1 * (Tpointer + 1))
            else Tpointer := (-1 * (Tpointer - 1));

           if not Found then hTrack := hTrack + Tpointer;
          until (Found or (hTrack = 71));
         end;
       end;
      Emu1581 : begin
        repeat
         hTrack := hTrack + Tpointer;
         hSector := 0;
         repeat
          Found := SectorIsFree(H, hTrack, hSector);
          if not Found then inc(hSector);
         until (Found or (hSector = 40));

         if (Device[bDev]^.Drv[bDrv].dTrack <> DIRtrack(bDev)) then
          begin
           if (Tpointer > 0) then Tpointer := (-1 * (Tpointer + 1))
            else Tpointer := (-1 * (Tpointer - 1));
          end;
        until (Found or (hTrack = MaxTracks(H)));
       end;
      EmuD16 : begin
        hTrack := 0;
        hSector := 0;

        repeat
         inc(hTrack);
         if (hTrack = DIRtrack(bDev)) then inc(hTrack);

         repeat
          Found := SectorIsFree(H, hTrack, hSector);
          if not Found then inc(hSector);
         until (Found or (hSector = 256));
        until (Found or (hTrack = MaxTracks(H)));
       end;
      Emu8050,
      Emu8250 : begin
        hTrack := 0;

        repeat
         inc(hTrack);
         if (hTrack = DIRtrack(bDev)) then inc(hTrack);
         hSector := 0;

         repeat
          Found := SectorIsFree(H, hTrack, hSector);

          if not Found then
           begin
            inc(hSector, 10);
            if (hSector > MaxSectors(H, hTrack)-1) then
              dec(hSector, MaxSectors(H, hTrack));
           end;
         until (Found or (hSector = 0));
        until (Found or (hTrack = MaxTracks(H)));
       end;
      Emu3040 : begin
        inc(hTrack);

        repeat
         hSector := 0;

         repeat
          Found := SectorIsFree(H, hTrack, hSector);

          if not Found then
           begin
            if (hTrack in [18..30]) then inc(hSector, 11)
              else inc(hSector, 10);
            if (hSector > MaxSectors(H, hTrack)-1) then
              dec(hSector, MaxSectors(H, hTrack));
           end;
         until (Found or (hSector = 0));

         if (Tpointer > 0) then Tpointer := (-1 * (Tpointer + 1))
          else Tpointer := (-1 * (Tpointer - 1));

         hTrack := hTrack + Tpointer;
        until (Found or (hTrack = 36));
       end;
     end;

    if Found then SetSector(H, hTrack, hSector);
   end;

  FindFreeSector := Found;
end;


function FType2txt(by : byte) : string;
{---------------------------------------------------------------------------
Translate filetype into readable text
---------------------------------------------------------------------------}
var pst : string[5];
begin
  case by of
    $01      : pst :='*SEQ ';
    $02      : pst :='*PRG ';
    $03      : pst :='*USR ';
    $05      : pst :='*DIR ';
    $80, $A0 : pst :=' DEL ';
    $81, $A1 : pst :=' SEQ ';
    $82, $A2 : pst :=' PRG ';
    $83, $A3 : pst :=' USR ';
    $84      : pst :=' REL ';
    $85,
    $86      : pst :=' DIR ';
    $C0      : pst :=' DEL<';
    $C1      : pst :=' SEQ<';
    $C2      : pst :=' PRG<';
    $C3      : pst :=' USR<';
    $C4      : pst :=' REL<';
    $C5,
    $C6      : pst :=' DIR<';
     else pst := '?????';
   end;
  FType2txt := pst;
end;


function InfoTrack(H : integer) : byte;
{---------------------------------------------------------------------------
Outputs the first track with info, normally the first BAM sector
---------------------------------------------------------------------------}
begin
  case Device[DriveHandles[H].hbDevNo]^.FStype of
    Emu1541,
    Emu1571,
    Emu3040,
    Emu4040,
    EmuD16  : InfoTrack := 18;

    Emu8050,
    Emu8250 : InfoTrack := 39;

    Emu1581 : InfoTrack := 40;
   end;
end;


function Label2Filename(pst, ext : str16; Dir : str80) : str16;
var w : word;
begin
  pst := upper(copy(ml(pst, 8, True), 1, 8));

  for w := 1 to 8 do
   if not(pst[w] in ['A'..'Z', '0'..'9', '_', '-', '$']) then
    pst[w] := '_';

  w := 0;
  while exists(Dir + pst + '.' + ext) do
   begin
    if w <   17 then pst := copy(pst, 1, 7) + Dec2Hex(w, 1) else
    if w <  256 then pst := copy(pst, 1, 6) + Dec2Hex(w, 2) else
    if w < 4097 then pst := copy(pst, 1, 5) + Dec2Hex(w, 3) else
     pst := copy(pst, 1, 4) + Dec2Hex(w, 4);

    inc(w);
   end;

  Label2Filename := pst;
end;


function MaxBAMs(bDev : byte) : byte;
{---------------------------------------------------------------------------
Outputs the maximum number of BAMs for the given drive
---------------------------------------------------------------------------}
begin
  case Device[bDev]^.FStype of
    Emu1541 : MaxBAMs := 1;
    Emu1571 : MaxBAMs := 2;
    Emu1581 : MaxBAMs := 2;
    Emu3040 : MaxBAMs := 1;
    Emu4040 : MaxBAMs := 1;
    Emu8050 : MaxBAMs := 2;
    Emu8250 : MaxBAMs := 4;
    EmuD16  : MaxBAMs := 8;
   end;
end;


function MaxSectors(H, Track : Byte) : Byte;
{---------------------------------------------------------------------------
Outputs the maximum number of sectors for the given drive and track
---------------------------------------------------------------------------}
begin
 with DriveHandles[H] do
  if Device[hbDevNo]^.FSType in [Emu1541, Emu4040] then
   Case Track of
     1..17 : MaxSectors := 21;
    18..24 : MaxSectors := 19;
    25..30 : MaxSectors := 18;
    31..35 : MaxSectors := 17;
   end
    else
  if Device[hbDevNo]^.FSType = Emu3040 then
   Case Track of
     1..17 : MaxSectors := 21;
    18..24 : MaxSectors := 20;
    25..30 : MaxSectors := 18;
    31..35 : MaxSectors := 17;
   end
    else
  if Device[hbDevNo]^.FSType = Emu1571 then
   Case Track of
     1..17, 36..52 : MaxSectors := 21;
    18..24, 53..59 : MaxSectors := 19;
    25..30, 60..65 : MaxSectors := 18;
    31..35, 66..70 : MaxSectors := 17;
   end
    else
  if Device[hbDevNo]^.FSType = Emu8050 then
   Case Track of
     1..39 : MaxSectors := 29;
    40..53 : MaxSectors := 27;
    54..64 : MaxSectors := 25;
    65..77 : MaxSectors := 23;
   end
    else
  if Device[hbDevNo]^.FSType = Emu8250 then
   Case Track of
     1..39,  78..116 : MaxSectors := 29;
    40..53, 117..130 : MaxSectors := 27;
    54..64, 131..141 : MaxSectors := 25;
    65..77, 142..154 : MaxSectors := 23;
   end
    else
  if Device[hbDevNo]^.FSType = Emu1581 then MaxSectors := 40
    else
  if Device[hbDevNo]^.FSType = EmuD16 then MaxSectors := 255;  { = 256 }
end;


function MaxTracks(H : integer) : Byte;
{---------------------------------------------------------------------------
Outputs the maximum number of tracks for the given drive
---------------------------------------------------------------------------}
begin
 with DriveHandles[H] do
  case Device[hbDevNo]^.FSType of
    Emu3040,
    Emu4040,
    Emu1541 : MaxTracks := 35;

    Emu1571 : MaxTracks := 70;

    Emu1581 : with Device[hbDevNo]^.Drv[hbDrvNo] do
     begin
      if (dTrack = DIRtrack(hbDevNo)) then MaxTracks := 80
       else MaxTracks := dTrack + dSize - 1;
     end;

    Emu8050 : MaxTracks := 77;

    Emu8250 : MaxTracks := 154;

    EmuD16  : MaxTracks := 255;
   end;
end;


function NoAccess(H : integer) : boolean;
{---------------------------------------------------------------------------
Checks if drive is valid or disk is locked due to a incorrect identifier
---------------------------------------------------------------------------}
var
  pbo : boolean;
  Emu : TEmutype;
begin
  NoAccess := True;

  with DriveHandles[H] do
  if not Device[hbDevNo]^.drv[hbDrvNo].InUse then
   begin
    DriveMsg(H, 74, 0, 0);
    exit;
   end;


{  Check diskformat  (disk can be locked)  }
  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    if dNoAccess then DriveMsg(H, 75, 40, 0)    { disk locked }
     else NoAccess := False;
end;


function SectorIsFree(H, fTrack, fSector : byte) : Boolean;
{----------------------------------------------------------------------------
Checks if an particular sector is unused
in:     track
        sector
out:    yes / no
----------------------------------------------------------------------------}
Var
  BAMpos : TBAMpos;
begin
  SeekBAMentry(H, fTrack, fSector, BAMpos);

  with BAMpos do
   begin
    ReadSector(H, Track, Sector);

    with DriveHandles[H] do
     with Device[hbDevNo]^.Drv[hbDrvNo] do
      SectorIsFree := (dDataBlock[bByt] and bBit > 0);
   end;
end;


procedure SeekBAMentry(H, bT, bS : byte; var BAMpos : TBAMpos);
{-----------------------------------------------------------------------------
Searches for exact bit in BAM for given drive, track and sector
In:  bTrack   track
     bSector  sector
Out: bTrack   track of BAM entry
     bSector  sector of BAM entry
     bByt     byte
     bBit     bit
-----------------------------------------------------------------------------}
begin
  BAMpos.Track := 18;
  BAMpos.Sector := 0;
  BAMpos.bBit := Power2[bS mod 8];

  with DriveHandles[H] do
   case Device[hbDevNo]^.FSType of
    Emu3040,
    Emu4040,
    Emu1541,
    Emu1571 : BAMpos.bByt := 1 + (bT * 4) + (bS div 8);
    Emu1581 : begin
      BAMpos.Track := Device[hbDevNo]^.Drv[hbDrvNo].dTrack;
      BAMpos.Sector := 1 + ((bT - 1) div 40);
      BAMpos.bByt := 17 + ((bT - 1) mod 40) * 6 + (bS div 8);
     end;
    Emu8050,
    Emu8250 : begin
      BAMpos.Track := 38;
      BAMpos.Sector := ((bT - 1) div 50) * 3;
      BAMpos.bByt := 7 + ((bT - 1) mod 50) * 5 + (bS div 8);
     end;
    EmuD16  : begin
      BAMpos.Sector := 224 + ((bT - 1) div 8);
      BAMpos.bByt := ((bT - 1) mod 8) * 32 + (bS div 8);
     end;
   end;
end;


function SlotOfs(Slot : Byte) : Byte;
{----------------------------------------------------------------------------
Calculates at which byte the entry of a record in a directoryblock starts
----------------------------------------------------------------------------}
begin
  SlotOfs := $02 + (Slot * $20);
end;


function TalkSec(H : integer) : Byte;
Var
  EOI  : Boolean;
  B    : Byte;
  ECnt : Word;
begin
  TalkSec := ByteIn(H, EOI);
end;


function TrackFree(H, I : byte) : Longint;
{----------------------------------------------------------------------------
Calculates number of free blocks on a track.
in:     track
out:    number of free blocks
----------------------------------------------------------------------------}
Var
  B, J, K : Byte;
  Free : Longint;
begin
  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
  case Device[hbDevNo]^.FSType of
    Emu3040,
    Emu4040,
    Emu1541 : begin
      ReadSector(H, BAMtrack(H, 0), BAMsector(H, 0));
      Free := dDataBlock[I * 4];
     end;
    Emu1571 : begin
      ReadSector(H, BAMtrack(H, 0), BAMsector(H, 0));
      if (I < 36) then Free := dDataBlock[I * 4]
       else Free := dDataBlock[$DD + I - 35];
     end;
    Emu1581 : begin
      dec(I);
      J := I div 40;
      I := I mod 40;
      ReadSector(H, BAMtrack(H, J), BAMsector(H, J));
      Free :=  dDataBlock[16 + (I * 6)];
     end;
    Emu8250,
    Emu8050 : begin
      dec(I);
      J := I div 50;
      I := I mod 50;
      ReadSector(H, BAMtrack(H, J), BAMsector(H, J));
      Free :=  dDataBlock[5 + (I * 5)];
     end;
    EmuD16  : begin
      dec(I);
      J := I div 8;
      I := I mod 8;
      ReadSector(H, BAMtrack(H, J), BAMsector(H, J));
      Free := 0;
      for J := 0 to 31 do
       begin
        B := dDataBlock[I * 32 + J];
        for K := 0 to 7 do
         if (B and Power2[K] > 0) then inc(Free);
       end;
     end;
   end;

  TrackFree := Free;
end;


function Translate2AbsPos(H : integer; Track, Sector : Longint) : Longint;
{----------------------------------------------------------------------------
Translates Track/Sector to liniar position in flatfile

D16:  Track 1..255, Sector 255

D30:  Track 1..35  Sector 17..21
      21 Sectors for track 1-17
      20 Sectors for track 18-24
      18 Sectors for track 25-30
      17 Sectors for track 31-35

D64:  Track 1..35  Sector 17..21
      21 Sectors for track 1-17
      19 Sectors for track 18-24
      18 Sectors for track 25-30
      17 Sectors for track 31-35

D71:  Track 1..70  Sector 17..21
      21 Sectors for track 1-17
      19 Sectors for track 18-24
      18 Sectors for track 25-30
      17 Sectors for track 31-35
      21 Sectors for track 36-52
      19 Sectors for track 53-49
      18 Sectors for track 50-65
      17 Sectors for track 66-70

D80:  Track 1..77  Sector 23..29
      29 Sectors for track 1-30
      27 Sectors for track 40-53
      25 Sectors for track 54-64
      23 Sectors for track 65-77

D81:  Track 1..80, Sector 0..39

D82:  Track 1..154  Sector 23..29
      29 Sectors for track 1-30
      27 Sectors for track 40-53
      25 Sectors for track 54-64
      23 Sectors for track 65-77
      29 Sectors for track 78-116
      27 Sectors for track 117-130
      25 Sectors for track 131-141
      23 Sectors for track 142-154

in:     track
        sector
out:    position
----------------------------------------------------------------------------}
Var
  TmpPos : Longint;
  CType : TEmutype;
begin

{ Check for valid track and sector }
  Translate2AbsPos := -1;
  if (Track > MaxTracks(H)) or (Track = 0) then exit;
  if (Sector > MaxSectors(H, Track)) then exit;

  TmpPos := 0;
  CType := Device[DriveHandles[H].hbDevNo]^.FSType;

  if (CType in [Emu1541, Emu4040]) then         { D64 }
   begin
    if Track > 17 then TmpPos := TmpPos + (17 * 21 * 256);
    if Track > 24 then TmpPos := TmpPos + (7 * 19 * 256);
    if Track > 30 then TmpPos := TmpPos + (6 * 18 * 256);

    Case Track of
       1..17 : TmpPos := TmpPos + (Pred(Track) * 21 * 256);
      18..24 : TmpPos := TmpPos + ((Pred(Track)-17) * 19 * 256);
      25..30 : TmpPos := TmpPos + ((Pred(Track)-24) * 18 * 256);
      31..35 : TmpPos := TmpPos + ((Pred(Track)-30) * 17 * 256);
     end;
   end
    else
  if (CType = Emu1571) then                           { D71 }
   begin
    if Track > 17 then TmpPos := TmpPos + (17 * 21 * 256);
    if Track > 24 then TmpPos := TmpPos + (7 * 19 * 256);
    if Track > 30 then TmpPos := TmpPos + (6 * 18 * 256);
    if Track > 35 then TmpPos := TmpPos + (5 * 17 * 256);
    if Track > 52 then TmpPos := TmpPos + (17 * 21 * 256);
    if Track > 59 then TmpPos := TmpPos + (7 * 19 * 256);
    if Track > 65 then TmpPos := TmpPos + (6 * 18 * 256);

    Case Track of
       1..17 : TmpPos := TmpPos + (Pred(Track) * 21 * 256);
      18..24 : TmpPos := TmpPos + ((Pred(Track)-17) * 19 * 256);
      25..30 : TmpPos := TmpPos + ((Pred(Track)-24) * 18 * 256);
      31..35 : TmpPos := TmpPos + ((Pred(Track)-30) * 17 * 256);
      36..52 : TmpPos := TmpPos + ((Pred(Track)-35) * 21 * 256);
      53..59 : TmpPos := TmpPos + ((Pred(Track)-52) * 19 * 256);
      60..65 : TmpPos := TmpPos + ((Pred(Track)-59) * 18 * 256);
      66..70 : TmpPos := TmpPos + ((Pred(Track)-65) * 17 * 256);
     end;
   end
    else
  if (CType = Emu1581) then TmpPos := (Track-1) * 40 * 256    { D81 }
    else
  if (CType = Emu8050) then                     { D80 }
   begin
    if Track > 39 then TmpPos := TmpPos + (39 * 29 * 256);
    if Track > 53 then TmpPos := TmpPos + (14 * 27 * 256);
    if Track > 64 then TmpPos := TmpPos + (11 * 25 * 256);

    Case Track of
       1..39 : TmpPos := TmpPos + (Pred(Track) * 29 * 256);
      40..53 : TmpPos := TmpPos + ((Pred(Track)-39) * 27 * 256);
      54..64 : TmpPos := TmpPos + ((Pred(Track)-53) * 25 * 256);
      65..77 : TmpPos := TmpPos + ((Pred(Track)-64) * 23 * 256);
     end;
   end
    else
  if (CType = Emu8250) then                     { D82 }
   begin
    if Track >  39 then TmpPos := TmpPos + (39 * 29 * 256);
    if Track >  53 then TmpPos := TmpPos + (14 * 27 * 256);
    if Track >  64 then TmpPos := TmpPos + (11 * 25 * 256);
    if Track >  77 then TmpPos := TmpPos + (13 * 23 * 256);
    if Track > 116 then TmpPos := TmpPos + (39 * 29 * 256);
    if Track > 130 then TmpPos := TmpPos + (14 * 27 * 256);
    if Track > 141 then TmpPos := TmpPos + (11 * 25 * 256);

    Case Track of
        1.. 39 : TmpPos := TmpPos + (Pred(Track) * 29 * 256);
       40.. 53 : TmpPos := TmpPos + ((Pred(Track)- 39) * 27 * 256);
       54.. 64 : TmpPos := TmpPos + ((Pred(Track)- 53) * 25 * 256);
       65.. 77 : TmpPos := TmpPos + ((Pred(Track)- 64) * 23 * 256);
       78..116 : TmpPos := TmpPos + ((Pred(Track)- 77) * 29 * 256);
      117..130 : TmpPos := TmpPos + ((Pred(Track)-116) * 27 * 256);
      131..141 : TmpPos := TmpPos + ((Pred(Track)-130) * 25 * 256);
      142..154 : TmpPos := TmpPos + ((Pred(Track)-141) * 23 * 256);
     end;
   end
    else
  if (CType = EmuD16) then TmpPos := (Track-1) * 256 * 256   { D16 }
    else
  if (CType = Emu3040) then                           { D30 }
   begin
    if Track > 17 then TmpPos := TmpPos + (17 * 21 * 256);
    if Track > 24 then TmpPos := TmpPos + (7 * 20 * 256);
    if Track > 30 then TmpPos := TmpPos + (6 * 19 * 256);

    Case Track of
      1..17 : TmpPos := TmpPos + (Pred(Track) * 21 * 256);
     18..24 : TmpPos := TmpPos + ((Pred(Track)-17) * 20 * 256);
     25..30 : TmpPos := TmpPos + ((Pred(Track)-24) * 18 * 256);
     31..35 : TmpPos := TmpPos + ((Pred(Track)-30) * 17 * 256);
    end;
   end
  else TmpPos := -1;

  TmpPos := TmpPos + Sector * 256;
  Translate2AbsPos := TmpPos;
end;


function Txt2FType(s : Str16) : byte;
{---------------------------------------------------------------------------
Translate text into filetype
---------------------------------------------------------------------------}
var pst : string[5];
    b : byte;
begin
  s := upper(s);
  pst := copy(s, 1, 3);

  if pst = 'DEL' then b := $80 else
  if pst = 'SEQ' then b := $81 else
  if pst = 'PRG' then b := $82 else
  if pst = 'USR' then b := $83 else
  if pst = 'DIR' then b := $86;

  if s[4] = '<' then b := b or $40;

  Txt2Ftype := b;
end;


function WildcardMatch(FName, Card : str80) : Boolean;
{-----------------------------------------------------------------------------
Determines if two filenames do match.
in:     Fname = Full filename
        Card  = Name with wildcards
out:    yes/no
-----------------------------------------------------------------------------}
Var
  bs, bf : array[0..7] of byte;
  bq, by, b2 : byte;
  pst, pst2 : str80;

  C    : Char;
  N, L : Byte;
begin
  WildcardMatch := False;
  bs[0] := pos('*', Card);
  bq := pos('?', Card);
  L  := Length(FName);

  if bs[0] = 0 then
   begin

{  Replace all '?' by character  }
    while ((bq > 0) and (bq < length(Card))) do
     begin
      Card[bq] := FName[bq];
      bq := pos('?', Card);
     end;
    WildcardMatch := (FName = Card);
    exit;
   end;

{  Remove all ?*, ** and *? combinations  }
  by := pos('?*', Card);
  while by > 0 do
   begin
    Card := copy(Card, 1, by-1) + copy(Card, by+1, length(Card)-1);
    by := pos('?*', Card);
   end;
  by := pos('*?', Card);
  while by > 0 do
   begin
    Card := copy(Card, 1, by) + copy(Card, by+2, length(Card)-1);
    by := pos('*?', Card);
   end;
  by := pos('**', Card);
  while by > 0 do
   begin
    Card := copy(Card, 1, by-1) + copy(Card, by+1, length(Card)-1);
    by := pos('**', Card);
   end;

{  Test for special case first  }
  if FName = '*' then
   begin
    WildcardMatch := True;
    exit;
   end;

{  Find out if there are more * wildcards, maximal 8  }
  bs[0] := pos('*', Card);
  pst := Card;
  for by := 1 to 7 do bs[by] := 0;
  for by := 1 to 7 do
   begin
    delete(pst, 1, bs[by-1]);
    bs[by] := pos('*', pst);
    if bs[by] = 0 then break;
   end;
  for by := 1 to 7 do if bs[by] > 0 then bs[by] := bs[by] + bs[by-1];

{  Does part fit in FName?  }
  WildcardMatch := False;
  bf[0] := 1;
  if bs[0] > 1 then                             { *xxxxx }
   begin
    pst := copy(FName, 1, bs[0]-1);
    pst2 := copy(card, 1, bs[0]-1);
    for by := 1 to length(pst) do if pst2[by] = '?' then pst[by] := '?';
    if pst2 <> pst then exit;
    bf[0] := length(pst) + 1;
   end;

  for N := 1 to 7 do
   begin
    if bs[N-1] = length(Card) then
     begin                                      { xxxxx* }
      WildcardMatch := True;
      exit;                                     { found }
     end;
    if bf[N-1] > L then exit;
    if bs[N] > 0 then
     begin
      pst2 := copy(Card, bs[N-1]+1, bs[N]-bs[N-1]-1);
      for b2 := bf[N-1] to L-length(pst2)+1 do
       begin
        pst := copy(FName, b2, length(pst2));
        for by := 1 to length(pst) do if pst2[by] = '?' then pst[by] := '?';
        if pst = pst2 then
         begin
          bf[N] := b2 + length(pst);
          break;                                { next combination }
         end;
        if b2 = L-length(pst2)+1 then exit;     { not found }
       end;
     end
      else
     begin
      pst2 := copy(Card, bs[N-1]+1, length(Card)-bs[N-1]);
      pst := copy(FName, L-length(pst2)+1, length(pst2));
      for by := 1 to length(pst) do if pst2[by] = '?' then pst[by] := '?';
      if pst2 = pst then WildcardMatch := True;
      exit;
     end;                                       { (not) found }
   end;
end;


{============================================================================}

function AtnFound : boolean;
begin
  AtnFound := (Port[wIeeeLptStat] and $40 = 0);
end;


function DavFound : boolean;
begin
  case bIeeeType of
   0,1,5 : DavFound := (Port[wIeeeLptCtrl] and 8 > 0);
   2     : DavFound := (Port[wIeeeLptStat] and $20 = 0);
  end;
end;


function EoiFound : boolean;
begin
  case bIeeeType of
   0,1,5 : EoiFound := (Port[wIeeeLptCtrl] and 1 > 0);
   2     : EoiFound := (Port[wIeeeLptStat] and $80 > 0);
  end;
end;


function NdacFound : boolean;
begin
  case bIeeeType of
   0,1,5 : NdacFound := (Port[wIeeeLptCtrl] and 2 > 0);
   2     : NdacFound := (Port[wIeeeLptStat] and 8 = 0);
  end;
end;


function NrfdFound : boolean;
begin
  case bIeeeType of
   0,1,5 : NrfdFound := (Port[wIeeeLptCtrl] and 4 = 0);
   2     : NrfdFound := (Port[wIeeeLptStat] and $10 = 0);
  end;
end;


function ResetFound : boolean;
begin
  ResetFound := (Port[wIeeeLptStat] and $40 = 0);
end;


procedure AtnHiIEEE;
begin
  case bIeeeType of
   0,2 : port[wIeeeComAddr + 4] := port[wIeeeComAddr + 4] or 1;
   1   : port[wIeeeComAddr + 4] := port[wIeeeComAddr + 4] and $FE;
   5   : begin
     bSerial2 := bSerial2 and $FE;
     port[wIeeeComAddr] := bSerial2;
    end;
  end;
end;


procedure AtnLoIEEE;
begin
  case bIeeeType of
   0,2 : port[wIeeeComAddr + 4] := port[wIeeeComAddr + 4] and $FE;
   1   : port[wIeeeComAddr + 4] := port[wIeeeComAddr + 4] or 1;
   5   : begin
     bSerial2 := bSerial2 or 1;
     port[wIeeeComAddr] := bSerial2;
    end;
  end;
end;


procedure AtnTrapIeeeOff;
begin
  case bIeeeType of
   0   : port[wIeeeComAddr + 4] := port[wIeeeComAddr + 4] and $FD;
   1,2 : port[wIeeeComAddr + 4] := port[wIeeeComAddr + 4] or 2;
   5   : begin
     bSerial2 := bSerial2 or 8;
     port[wIeeeComAddr] := bSerial2;
    end;
  end;
end;


procedure AtnTrapIeeeOn;
begin
  case bIeeeType of
   0   : port[wIeeeComAddr + 4] := port[wIeeeComAddr + 4] or 2;
   1,2 : port[wIeeeComAddr + 4] := port[wIeeeComAddr + 4] and $FD;
   5   : begin
     bSerial2 := bSerial2 and $F7;
     port[wIeeeComAddr] := bSerial2;
    end;
  end;
end;


procedure DavHi;
begin
  bSerial := bSerial and $F7;
  port[wIeeeLptCtrl] := bSerial;
end;


procedure DavLo;
begin
  bSerial := bSerial or 8;
  port[wIeeeLptCtrl] := bSerial;
end;


procedure EoiHi;
begin
  bSerial := bSerial and $FE;
  port[wIeeeLptCtrl] := bSerial;
end;


procedure EoiLo;
begin
  bSerial := bSerial or 1;
  port[wIeeeLptCtrl] := bSerial;
end;


procedure InitIEEE;
{-----------------------------------------------------------------------------
 initialize IEEE
-----------------------------------------------------------------------------}
begin
  bSerial := $E4;
  bSerial2 := $EC;
  RenHi;
  ResetHiIEEE;
  AtnTrapIeeeOff;
  AtnHiIEEE;
  DavHi;
  EoiHi;
  NrfdHi;
  NdacHi;
end;


procedure LedOff;
begin
  NoSound;
end;


procedure LedOn;
begin
  sound(1000);
end;


procedure NdacHi;
begin
  bSerial := bSerial and $FD;
  port[wIeeeLptCtrl] := bSerial;
end;


procedure NdacLo;
begin
  bSerial := bSerial or 2;
  port[wIeeeLptCtrl] := bSerial;
end;


procedure NrfdHi;
begin
  bSerial := bSerial or 4;
  port[wIeeeLptCtrl] := bSerial;
end;


procedure NrfdLo;
begin
  bSerial := bSerial and $FB;
  port[wIeeeLptCtrl] := bSerial;
end;


procedure ParallelInput;
{-----------------------------------------------------------------------------
Switch the parallel port of the hybrid/parallel cable to input
-----------------------------------------------------------------------------}
begin
  port[wIeeeLptAddr] := $FF;
  bSerial := bSerial or sbBidir;
  port[wIeeeLptCtrl] := bSerial;
end;


procedure ParallelOutput;
{---------------------------------------------------------------------------
Switch the parallel port of the hybrid/parallel cable to output
---------------------------------------------------------------------------}
begin
  bSerial := bSerial and (not sbBidir);
  port[wIeeeLptCtrl] := bSerial;
end;


procedure RenHi;
begin
  case bIeeeType of
   5   : begin
     bSerial2 := bSerial2 and $FD;
     port[wIeeeComAddr] := bSerial2;
    end;
  end;
end;


procedure RenLo;
begin
  case bIeeeType of
   5   : begin
     bSerial2 := bSerial2 or 2;
     port[wIeeeComAddr] := bSerial2;
    end;
  end;
end;


procedure ResetHiIEEE;
begin
  case bIeeeType of
   0 : port[wIeeeComAddr + 3] := port[wIeeeComAddr + 3] and $BF;
   2 : port[wIeeeComAddr + 3] := port[wIeeeComAddr + 3] or $40;
   5   : begin
     bSerial2 := bSerial2 or 4;
     port[wIeeeComAddr] := bSerial2;
    end;
  end;
end;


procedure ResetLoIEEE;
begin
  AtnTrapIeeeOff;  { to make sure at least one COM line supplies power }

  case bIeeeType of
   0 : port[wIeeeComAddr + 3] := port[wIeeeComAddr + 3] or $40;
   2 : port[wIeeeComAddr + 3] := port[wIeeeComAddr + 3] and $BF;
   5   : begin
     bSerial2 := bSerial2 and $FB;
     port[wIeeeComAddr] := bSerial2;
    end;
  end;
end;


{============================================================================}


procedure AddSector(H : integer);
var
  K : integer;
  DirEntryDxx : TDirEntryDxx;
begin
  K := AssignHandle(H, 'XXX');
  DriveHandles[K] := DriveHandles[H];

  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    begin
{*  Allocate a new sector  *}
     if not FindFreeSector(K) then
      begin
       DiskStatus := -3;
       DriveMsg(H, 72, 0, 0);
       exit;
      end;

{*  Put it in the chain  *}
     ReadSector(H, hTrack, hSector);
     dDataBlock[0] := DriveHandles[K].hTrack;
     dDataBlock[1] := DriveHandles[K].hSector;
     WriteSector(H, hTrack, hSector);

     GetDirEntry(H, DirEntryDxx);

{*  Update size  *}
     Inc(DirEntryDxx.Lo_Size);
     if DirEntryDxx.Lo_Size = 0 then Inc(DirEntryDxx.Hi_Size);
     PutDirEntry(H, DirEntryDxx);

{*  New info for write  *}
     hTrack  := DriveHandles[K].hTrack;
     hSector := DriveHandles[K].hSector;
     hOffset := 2;
    end;

  CloseHandle(K);
end;


procedure ByteOut(H, Value : Byte; DoEOI : Boolean);
Var
  pbo : boolean;
  B, I : Byte;
  ECnt : word;
label L1, L2, L3;
begin
  if BoTest then exit;


  Error := 0;
  AtnTrapIeeeOn;
  boEOI := DoEOI;
  boATNfound := False;

  NdacHi;
  NrfdHi;
  if AtnFound then goto L1;           { ATN = (L), -> }

{ Wait for NRFD = (H) }
  ECnt := 40000;

  while NrfdFound do
   begin
    if AtnFound then goto L1;         { ATN = (L), -> }

{##    dec(ECnt);  }

    if ECnt = 0 then
     begin
write('TO1 ');
      goto L2;                        { timeout, -> }
     end;
   end;

  ParallelOutput;
  Port[wIeeeLptAddr] := not(Value);
  if boEOI then EoiLo;

{ Wait for NDAC = (L) }
  ECnt := 40000;

  while not NdacFound do
   begin
    if AtnFound then goto L1;         { ATN = (L), -> }

{##    dec(ECnt);}

    if ECnt = 0 then
     begin
write('TO2 ');
      goto L2;                        { timeout, -> }
     end;
   end;

  DavLo;

{Wait for NDAC = (H)}
  ECnt := 40000;

  while NdacFound do
   begin

    if AtnFound then goto L1;         { ATN = (L), -> }

{##    dec(ECnt);}

    if ECnt = 0 then
     begin
write('TO3 ');
      goto L2;                     { timeout, -> }
     end;
   end;

L3:
  ParallelInput;
  AtnTrapIeeeOn;
  NdacHi;
  NrfdHi;
  DavHi;
  EoiHi;
  exit;

{ ATN = (L) }
L1:
  NrfdLo;
  NdacLo;
  AtnTrapIeeeOff;
  boATNfound := True;
  exit;

L2:
  Error := $81;
write(' E81 ');
  goto L3;
end;


procedure ByteOut2(H, Value : Byte; DoEOI : Boolean);
Var
  pbo : boolean;
  B, I : Byte;
  ECnt : word;
label L2, L3;
begin

asm
   { cli }
end;

  Error := 0;
  boEOI := DoEOI;
  boATNfound := False;
  NdacHi;
  NrfdHi;

{ Wait for NRFD = (H) }
  ECnt := 40000;
  while NrfdFound do
   begin
    dec(ECnt);
    if ECnt = 0 then
     begin
      goto L2;                     { timeout, -> }
     end;
   end;

  ParallelOutput;
  Port[wIeeeLptAddr] := not(Value);
  if boEOI then EoiLo;

{ Wait for NDAC = (L) }
  ECnt := 40000;
  while not NdacFound do
   begin
    dec(ECnt);
    if ECnt = 0 then
     begin
      goto L2;                     { timeout, -> }
     end;
   end;

  DavLo;

{Wait for NDAC = (H)}
  ECnt := 40000;
  while NdacFound do
   begin
    dec(ECnt);
    if ECnt = 0 then
     begin
      goto L2;                     { timeout, -> }
     end;
   end;

L3:
  ParallelInput;
  NdacHi;
  NrfdHi;
  DavHi;
  EoiHi;
  exit;

L2:
  Error := $81;
write(' E81 ');
  goto L3;
end;


procedure ChangeDirectory(H : integer);
var
  b : byte;
  s, s2, s3 : str80;
  s5 : string[5];
  ft : text;
begin
  with DriveHandles[H] do with Device[hbDevNo]^.Drv[hbDrvNo] do
   begin
    hFileType := $86;

    if Device[hbDevNo]^.FStype = EmuPC then
     begin
      ChDir(sDirPath);
      if not exists(stCBMdir) then
       begin
        DriveMsg(H, 62, 0, 2);
        ChDir(OrgDir);
        exit;
       end;

      assign(ft, stCBMdir);
      reset(ft);

      while not EOF(ft) do
       begin
        readln(ft, s);
        s3 := short(copy(s, 1, 40));  { DOS name }
        delete(s, 1, 41);
        s2 := short(copy(s, 1, 16));  { CBM name }
        delete(s, 1, 17);
        s5 := copy(s, 1, 3);          { CBM type }

        if (s5 = 'DIR') and (s2 = hFileName) then
         begin
          {$I-}
          ChDir(hFileName);
          if IOresult > 0 then DriveMsg(H, 47, IOresult, 0);
          {$I+}
          Break;
         end
       end;

      close(ft);
      getdir(0, sDirPath);
      ChDir(OrgDir);
     end
      else

    if Device[hbDevNo]^.FStype = EmuPC then
     begin
      ChDir(sDirPath);

      {$I-}
      ChDir(hFileName);
      if IOresult > 0 then DriveMsg(H, 47, IOresult, 0);
      {$I+}

      getdir(0, sDirPath);
      ChDir(OrgDir);
     end
    else
     begin
      if hFileName[1] = chCsep then
       begin
        if (length(hFileName) = 1) then         { goto ROOT }
         begin
          dSector := DIRsector(hbDevNo);
          dTrack := DIRtrack(hbDevNo);
          sDirPath := chCsep;
         end
        else DriveMsg(H, 30, 0, 0);
       end
        else

      if hFileName = '..' then
       begin
        b := RightPos(chCsep, sDirPath);
        if b > 0 then
         begin
          hFiletype := $C6;
          FindFile(H);
          dSector := hsSector;
          dTrack := hsTrack;
          sDirPath[0] := chr(b-1);
          if sDirPath[0] = #0 then sDirPath := chCsep;
         end;
       end
        else
      if not FindFile(H) then DriveMsg(H, 46, 0, 0) else
       begin
        dSector := hSector;
        dTrack := hTrack;

        if sDirPath[length(sDirPath)] <> chCsep then
         sDirPath := sDirPath + chCsep;
        sDirPath := sDirPath + hFileName;
       end;
     end;
   end;
end;


procedure CloseAllHandles(bDevNo, bDrvNo : Integer);
{---------------------------------------------------------------------------
Closes all Hs of this particular drive
---------------------------------------------------------------------------}
var b1, b2 : integer;
begin
  for b1 := 1 to coMaxDriveHandles do
   with DriveHandles[b1] do
    begin
     b2 := b1;
     if ((hbDevNo = bDevNo) and (hbDrvNo = bDrvNo)) then CloseHandle(b2);
    end;

  for b1 := 0 to 15 do
   Device[bDevNo]^.Channel[b1].cHandle := 0;
end;


procedure CloseHandle(var H : Integer);
begin
  with DriveHandles[H] do Device[hbDevNo]^.Drv[hbDrvNo].DiskStatus := 0;
  Fillchar(DriveHandles[H], Sizeof(DriveHandles[H]), 0);
  H := 0;
end;


procedure CreateSubdirD81(H : integer; Name : str16; fTrack : byte; fNumSectors : word);
Var
  B, C : Byte;
  DirEntryDxx : TDirEntryDxx;
label PL1;
begin
  if NoAccess(H) then exit;

  if (Pos('*', Name) + Pos('?', Name) > 0) then
   begin
    DriveMsg(H, 62, 0, 0);
    Exit;
   end;

{  First check if given sectors are still free  }
  for B := fTrack to (fTrack + (fNumSectors div 40) -1) do
   for C := 0 to 39 do
    if (not SectorIsFree(H, B, C)) then
     begin
      DriveMsg(H, 67, 0, 0);
      Exit;
     end;

{  Save subdirectory to disk  }
  H := AssignHandle(H, Name);
  if (H < 0) then exit;

  DriveHandles[H].hFiletype := $86;
  if not AddDirEntry(H) then exit;

  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    if (DiskStatus <> 0) then goto PL1;

  if not AllocateSector(H) then exit;


{  Allocate requested sectors for partition  }
  for B := fTrack to (fTrack + (fNumSectors div 40) -1) do
   for C := 0 to 39 do SetSector(H, B, C);

  GetDirEntry(H, DirEntryDxx);
  with DirEntryDxx do
   begin

{  Free original allocated sector  }
    ResetSector(H, Track_1, Sector_1);

{  Add given ones  }
    Track_1 := fTrack;
    Sector_1 := 0;
    Lo_Size  := fNumSectors mod 256;
    Hi_Size  := fNumSectors div 256;
   end;
  PutDirEntry(H, DirEntryDxx);

{  Creat BAM, directory etc.  }
  ReadSector(H, 40, 0);
  WriteSector(H, fTrack, 0);
  ReadSector(H, 40, 1);
  WriteSector(H, fTrack, 1);
  ReadSector(H, 40, 2);
  WriteSector(H, fTrack, 2);

  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    begin
     for B := 0 to 255 do dDataBlock[B] := 0;
     dDataBlock[1] := $FF;
    end;

  WriteSector(H, fTrack, 3);

PL1:
  CloseHandle(H);
end;


procedure DriveMsg(H, E, Track, Sector : word);
var
  pst : str32;
begin
  if bOldError = E then exit;
  Error := E;
  boError := True;
  str(E, pst);
  bOldError := Error;

  case E of
      0: sError := 'OK';
      1: sError := 'FILES SCRATCHED';

     20: sError := 'READ ERROR';
     21: sError := 'READ ERROR';
     22: sError := 'READ ERROR';
     23: sError := 'READ ERROR';
     24: sError := 'READ ERROR';
     25: sError := 'WRITE ERROR';
     26: sError := 'WRITE PROTECT ON';
     27: sError := 'READ ERROR';
     28: sError := 'WRITE ERROR';
     29: sError := 'DISK ID MISMATCH';
     30: sError := 'SYNTAX ERROR';
     31: sError := 'SYNTAX ERROR';
     32: sError := 'SYNTAX ERROR';
     33: sError := 'SYNTAX ERROR';
     34: sError := 'SYNTAX ERROR';
     39: sError := 'SYNTAX ERROR';
     40: sError := 'DISK LOCKED';
     41: sError := 'YET TO BE PROGRAMMED';
     43: sError := 'INVALID COMMAND';
     44: sError := 'NOT SUPPORTED';
     45: sError := 'DIRECTORY EXISTS';
     46: sError := 'DIRECTORY NOT FOUND';
     47: sError := 'DOS ERROR';
     48: sError := 'WRITE BYTE ERROR';
     49: sError := 'FORMAT ERROR';
     50: sError := 'RECORD NOT PRESENT';
     51: sError := 'OVERFLOW IN RECORD';
     52: sError := 'FILE TOO LARGE';
     60: sError := 'WRITE FILE OPEN';
     61: sError := 'FILE NOT OPEN';
     62: sError := 'FILE NOT FOUND';
     63: sError := 'FILE EXISTS';
     64: sError := 'FILE TYPE MISMATCH';
     65: sError := 'NO BLOCK';
     66: sError := 'ILLEGAL TRACK AND SECTOR';
     67: sError := 'ILLEGAL SYSTEM T OR S';
     70: sError := 'NO CHANNEL';
     71: sError := 'DIRECTORY ERROR';
     72: sError := 'DISK FULL';
     73: sError := 'CBM-HD V1.0';
     74: sError := 'DRIVE NOT READY';
     75: sError := 'FORMAT ERROR';
     76: sError := 'CONTROLLER ERROR';
     77: sError := 'SELECTED PARTITION ILLEGAL';
     78: sError := 'NO H AVAILABLE';
     79: sError := 'READ BYTE ERROR';
     80: sError := 'FILE IS LOCKED';
      else sError := '?????? ' + pst + ' ???????';
   end;

  with DriveHandles[H] do
   with Device[hbDevNo]^ do
    if E <> 1 then
     begin
      dErrorMsg := pst + ', ' + sError + ', ';
      str(Track, pst);
      dErrorMsg := dErrorMsg + ', ' + pst;
      str(Sector, pst);
      dErrorMsg := dErrorMsg + ', ' + pst + #13;
      if Drv[hbDrvNo].DiskStatus = 0 then Drv[hbDrvNo].DiskStatus := -1;
     end
    else
     begin
      str(Track, pst);
      dErrorMsg := pst + ', ' + sError + ', 0, 0';
     end;
end;


procedure FindFirstDxx(H : integer);
{---------------------------------------------------------------------------
Look for first entry of the root directory
---------------------------------------------------------------------------}
begin
  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    begin
     hdTrack := dTrack;
     hdSector := dSector;
     hSlot := 255;                   { FindNextDxx will make it point to 0 }
    end;
end;


procedure FindNextDxx(H : integer);
{---------------------------------------------------------------------------
Look for next valid file entry
---------------------------------------------------------------------------}
Var
  I : Byte;
  FFNTrack, FFNSector : byte;
  DirEntryDxx : TDirEntryDxx;
begin
  with DriveHandles[H] do with Device[hbDevNo]^.Drv[hbDrvNo] do
   begin
    DiskStatus := -1;
    GetDirEntry(H, DirEntryDxx);
    FFNTrack  := dDataBlock[0];
    FFNSector := dDataBlock[1];

    repeat
     if ((hSlot < 7) or (hSlot = 255)) then
      Inc(hSlot)
     else
      begin
       if FFNTrack = 0 then Exit;

       hdTrack  := FFNTrack;
       hdSector := FFNSector;
       hSlot   := 0;
      end;

     GetDirEntry(H, DirEntryDxx);
     FFNTrack  := Device[hbDevNo]^.Drv[hbDrvNo].dDataBlock[0];
     FFNSector := Device[hbDevNo]^.Drv[hbDrvNo].dDataBlock[1];
    until (DirEntryDxx.Filetype > 0);

    DiskStatus := 0;
   end;

  with DriveHandles[H] do with DirEntryDxx do
   begin
    hFilename := '';
    hFiletype := Filetype;

    for I := 0 to 15 do
     if FName[I] = #160 then Break else hFilename := hFilename + FName[I];

    hsTrack    := TSideSec;
    hsSector   := SSideSec;
    hRecLen    := RecLength;
    hTrack     := Track_1;
    hSector    := Sector_1;
    hBlockSize := (256*Hi_Size) + Lo_Size;
   end;
end;


procedure GCRdecode(var GCR : TGCRbytes);
const
  GCRar : array[0..31] of byte =
     ($0, $0, $0, $0, $0, $0, $0, $0, $0, $8, $0, $1, $0, $C, $4, $5,
      $0, $0, $2, $3, $0, $F, $6, $7, $0, $9, $A, $B, $0, $D, $E, $0);
var
  b1 : byte;
  GCRorg : TGCRbytes;
begin
  GCRorg := GCR;

  GCR.g1 := GCRar[(GCRorg.g1 and $F8) shr 3] shl 4;
  b1     := ((GCRorg.g1 and 3) shl 2) or ((GCRorg.g2 and $C0) shr 6);
  GCR.g1 := GCR.g1 or GCRar[b1];

  GCR.g2 := GCRar[(GCRorg.g2 and $3E) shr 1] shl 4;
  b1     := ((GCRorg.g2 and 1) shl 4) or ((GCRorg.g3 and $F0) shr 4);
  GCR.g2 := GCR.g2 or GCRar[b1];

  b1     := ((GCRorg.g3 and $0F) shl 1) or ((GCRorg.g4 and $80) shr 7);
  GCR.g3 := GCRar[b1] shl 4;
  GCR.g3 := GCR.g3 or GCRar[((GCRorg.g4 and $7C) shl 2)];

  b1     := ((GCRorg.g4 and $03) shl 3) or ((GCRorg.g5 and $E0) shr 5);
  GCR.g4 := GCRar[b1] shl 4;
  GCR.g4 := GCR.g4 or GCRar[GCRorg.g5 and $1F];
end;


procedure GCRencode(var GCR : TGCRbytes);
const
  GCRar : array[0..15] of byte =
     ($0A, $0B, $12, $13, $0E, $0F, $16, $17,
      $09, $19, $1A, $1B, $0D, $1D, $1E, $15);
var
  b1 : byte;
  GCRorg : TGCRbytes;
begin
  GCRorg := GCR;

  GCR.g1 := GCRar[GCRorg.g1 shr  4] shl 3;
  b1     := GCRar[GCRorg.g1 and 15] shr 2;
  GCR.g1 := GCR.g1 or b1;

  GCR.g2 := GCRar[GCRorg.g1 and 15] shl 6;
  b1     := GCRar[GCRorg.g2 shr  4] shl 1;
  GCR.g2 := GCR.g2 or b1;
  b1     := GCRar[GCRorg.g2 and 15] shr 4;
  GCR.g2 := GCR.g2 or b1;

  GCR.g3 := GCRar[GCRorg.g2 and 15] shl 4;
  b1     := GCRar[GCRorg.g3 shr  4] shr 1;
  GCR.g3 := GCR.g3 or b1;

  GCR.g4 := GCRar[GCRorg.g3 shr  4] shl 7;
  b1     := GCRar[GCRorg.g3 and 15] shl 2;
  GCR.g4 := GCR.g4 or b1;
  b1     := GCRar[GCRorg.g4 shr  4] shr 3;
  GCR.g4 := GCR.g4 or b1;

  GCR.g5 := GCRar[GCRorg.g4 shr  4] shl 5;
  b1     := GCRar[GCRorg.g4 and 15];
  GCR.g5 := GCR.g5 or b1;
end;


procedure GetDirEntry(H : integer; Var DirEntryDxx : TDirEntryDxx);
Var P : PDirEntryDxx;
begin
  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    begin
     ReadSector(H, hdTrack, hdSector);
     P := PDirEntryDxx(Ptr(Seg(dDataBlock), Ofs(dDataBlock) + SlotOfs(hSlot)));
    end;

  Move(P^, DirEntryDxx, Sizeof(DirEntryDxx));
end;


procedure InitDxx(H : integer);
{---------------------------------------------------------------------------
Initial read from the Dxx-file
---------------------------------------------------------------------------}
Var
  b : byte;
  c : char;
  DirInfo : SearchRec;
  s, pst : Str23;
  Emu : TEmutype;
begin
  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    if Device[hbDevNo]^.FSType in [EmuPC, EmuCBM] then
     begin
      {$I-}
      ChDir(chr(Device[hbDevNo]^.Drv[hbDrvNo].DosDrive + 64) + ':\');
      b := IOresult;
      {$I+}

      if b <> 0 then
       begin
        DriveMsg(H, 47, b, 0);
        Exit;
       end;

      FindFirst('\*.*', $08, DirInfo);

      if DosError = 0 then s := ml(DirInfo.Name, 16, True)
       else s := Blank(16);

      if s[9] = '.' then s := copy(s, 1, 8) + copy(s, 10, 7) + ' ';

      BAMlabel := s + '  PC '
                   + chr(Device[hbDevNo]^.Drv[hbDrvNo].DosDrive + 64) + ':';
      dSector := 0;
      dTrack := 0;
      DiskStatus := 0;
      ChDir(OrgDir);
      dNoAccess := False;                       { always access }
     end
    else
     begin

  {  Read first info sector  }
      LastSector := 0;
      LastTrack := 0;
      Emu := Device[hbDevNo]^.FSType;
      ReadSector(H, InfoTrack(H), 0);

      case Emu of
        EmuD16,
        Emu3040,
        Emu4040,
        Emu1541,
        Emu1571 : Move(dDataBlock[144], pst[1], 23);
        Emu1581 : Move(dDataBlock[4], pst[1], 23);
        Emu8050,
        Emu8250 : Move(dDataBlock[6], pst[1], 23);
       end;

  {  Check diskformat  (disk can be locked)  }
      dNoAccess := False;
      if (Emu in [Emu4040, Emu1541, Emu1571]) then
       dNoAccess := dDataBlock[2] <> ord('A')  else
      if (Emu in [Emu8050, Emu8250]) then
       dNoAccess := dDataBlock[2] <> ord('C')  else
      if (Emu = Emu1581) then dNoAccess := dDataBlock[2] <> ord('D')  else
      if (Emu = EmuD16) then dNoAccess := dDataBlock[2] <> ord('R');

      pst[0] := #23;
      BAMlabel := pst;
      dSector := DIRsector(hbDevNo);
      dTrack := DIRtrack(hbDevNo);
      DiskStatus := 0;
     end;
end;


procedure PutDirEntry(H : integer; Var DirEntryDxx : TDirEntryDxx);
begin
  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    begin
     ReadSector(H, hdTrack, hdSector);
     Move(DirEntryDxx, dDataBlock[SlotOfs(hSlot)], Sizeof(DirEntryDxx));
     WriteSector(H, hdTrack, hdSector);
    end;
end;


procedure ReadBlockDxx(H : Integer; ReadData : Word;
                        Var Data; Var DataRead : Word; var pbo : boolean);
Var
  T,
  S,
  RD       : Byte;              { Amount to read }
  P        : Pointer;
begin
  DataRead := 0;

{  do some initial stuff  }
  P := @Data;
  with DriveHandles[H] do ReadSector(H, hTrack, hSector);

  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    begin
     T := dDataBlock[0];
     S := dDataBlock[1];
    end;

  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
  while ReadData > 0 do
   begin

  {  Last sector has been read  }
    if (hOffset > 255) or ((T = 0) and (hOffset > S)) then
     begin
      if T = 0 then Exit;
      hOffset := 2;
      hTrack  := T;
      hSector := S;
      ReadSector(H, hTrack, hSector);
      T := dDataBlock[0];
      S := dDataBlock[1];
     end;

  {*  Get max we may read  *}
    if ReadData > 256-hOffset then RD := 256-hOffset else RD := ReadData;

  {*  Limit it to what is available  *}
    if ((T = 0) and (RD > S-1)) then RD := S - 1;
    pbo := ((T = 0) and (RD = S-1));

  {*  Update vars  *}
    ReadData := ReadData - RD;
    DataRead := DataRead + RD;

  {*  Move the data  *}
    Move(dDataBlock[hOffset], P^, RD);
    P := Pointer(Ptr(Seg(P^), Ofs(P^)+RD));

  {*  Update internals  *}
    hOffset  := hOffset + RD;
    hPosition := hPosition + RD;
   end;
end;


procedure ReadSector(H, T, S : byte);
{---------------------------------------------------------------------------
Locates the specified sector and reads its contents
---------------------------------------------------------------------------}
var
  DxxFile : File;
  i : integer;
  l : longint;
begin
{  Check if we can skip reading the sector  }
  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    begin
     if LastTrack <> 0 then
      if ((LastTrack = T) and (LastSector = S)) then exit;

     l := Translate2AbsPos(H, T, S);
{

if (t=18) and (s=1) then
 s := 1;
}
     if (l = -1) then
      begin
       for i := 0 to Sizeof(dDataBlock) do dDataBlock[i] := 0;
       LastTrack := 0;
       Error := 99;
       sError := '< 1 >';
       exit;
      end;

     Assign(DxxFile, Device[hbDevNo]^.Drv[hbDrvNo].PCfile);

     Reset(DxxFile, 1);
     Seek(DxxFile, l);

     {$I-} BlockRead(DxxFile, dDataBlock, Sizeof(dDataBlock)); {$I+}
     i := IOresult;
     if i<>0 then writeln('*RS ', T, ' ', S, ' ', l);

     Close(DxxFile);
     LastTrack := T;
     LastSector := S;
    end;
end;


procedure ResetSector(H, fTrack, fSector : byte);
{---------------------------------------------------------------------------
Resets the corresponding bit in the BAM table
---------------------------------------------------------------------------}
begin
  SubSector(H, fTrack, fSector, False);
end;


procedure ScratchDxx(H : integer);
Var
  S, T : Byte;
  fb : file;
  pCom : str80;
  DirEntryDxx : TDirEntryDxx;
  Attr : word;
begin
  with DriveHandles[H] do with Device[hbDevNo]^do with Drv[hbDrvNo] do
   if FStype = EmuPC then
    begin
     ChDir(sDirPath);
     pCom := hFilename + '.' + short(FType2txt(hFiletype));
     assign(fb, pCom);

     if hForceDel then
      begin
       swapvectors;
       exec(getenv('COMSPEC'),'/c attrib -r ' + pCom);
       swapvectors;
       s := DosError;
      end;

     GetFAttr(fb, Attr);
     if Attr and ReadOnly > 0 then
      begin
       DiskStatus := -4;                         { File locked }
       DriveMsg(H, 80, 0, 0);
       ChDir(OrgDir);
       exit;
      end;

     {$I-}
     erase(fb);
     s := IOresult;
     {$I+}
     if s <> 0 then
      begin
       DiskStatus := -4;
       DriveMsg(H, 47, s, 0);
      end;

     ChDir(OrgDir);
    end
   else
    begin
     GetDirEntry(H, DirEntryDxx);

{  Check if file is locked  }
     if (not hForceDel) and (hFiletype and $40 > 0) then
      begin
       DiskStatus := -4;                         { File locked }
       DriveMsg(H, 80, 0, 0);
       exit;
      end;

{  Entry exists so free all the sectors (except DEL-files)  }
     if (hFiletype and 7 > 0) then
      begin
       T := hTrack;
       S := hSector;

       while (T > 0) do
        begin
         ResetSector(H, T, S);
         ReadSector(H, T, S);
         T := dDataBlock[0];
         S := dDataBlock[1];
        end;

{  Relative files need extra care because of side sectors  }
       if (hFiletype = $84 ) then
        begin
         GetDirEntry(H, DirEntryDxx);
         T := DirEntryDxx.TSideSec;
         S := DirEntryDxx.SSideSec;

         while (T > 0) do
          begin
           ResetSector(H, T, S);
           ReadSector(H, T, S);
           T := dDataBlock[0];
           S := dDataBlock[1];
          end;
        end;
      end;

     DirEntryDxx.Filetype := 0;
     PutDirEntry(H, DirEntryDxx);
    end;
end;


procedure SetSector(H, fTrack, fSector : byte);
{---------------------------------------------------------------------------
Sets the corresponding bit in the BAM table
---------------------------------------------------------------------------}
begin
  SubSector(H, fTrack, fSector, True);
end;


procedure SubSector(H, fTrack, fSector : byte; pbo : boolean);
{---------------------------------------------------------------------------
(re-) sets the corresponding bit in the BAM table
 in:  pbo  True = set, False = reset
---------------------------------------------------------------------------}
var
  BAMpos : TBAMpos;
begin
  SeekBAMentry(H, fTrack, fSector, BAMpos);

  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    with BAMpos do
   begin
    ReadSector(H, Track, Sector);

    if pbo then dDataBlock[bByt] := dDataBlock[bByt] and (not bBit)
     else dDataBlock[bByt] := dDataBlock[bByt] or bBit;

    case Device[hbDevNo]^.FSType of
      Emu3040,
      Emu4040,
      Emu1541 : if pbo then dec(dDataBlock[fTrack * 4])
                 else inc(dDataBlock[fTrack * 4]);
      Emu1571 : if pbo then dec(dDataBlock[fTrack * 4])
                 else inc(dDataBlock[fTrack * 4]);
      Emu1581 : if pbo then dec(dDataBlock[10 + fTrack * 6])
                 else inc(dDataBlock[10 + fTrack * 6]);
      Emu8250,
      Emu8050 : if pbo then dec(dDataBlock[1 + fTrack * 5])
                 else inc(dDataBlock[1 + fTrack * 5]);
     end;

    WriteSector(H, Track, Sector);
   end;
end;


procedure Validate (H : integer);
var
  ADir : ^TADir;
  tBAM : array[0..7] of Block256;
  pbo : boolean;
  b1, b2, bByte, bBit, bBAM : byte;
  DirTel, lADir : longint;
  BAMpos : TBAMpos;
  Emu : TEmutype;

 procedure SeekBAM;
 begin
   bBit := Power2[b2 mod 8];

   with DriveHandles[H] do with Device[hbDevNo]^ do
   case FSType of
    Emu3040,
    Emu4040,
    Emu1541,
    Emu1571 : begin
      bByte := 1 + (b1 * 4) + (b2 div 8);
      bBam := 0;
     end;
    Emu8050,
    Emu8250 : begin
      bBAM := (b1 - 1) div 50;
      bByte := 7 + ((b1 - 1) mod 50) * 5 + (b2 div 8);
     end;
    EmuD16  : begin
      bBAM := ((b1 - 1) div 8);
      bByte := ((b1 - 1) mod 8) * 32 + (b2 div 8);
     end;
   end;
 end;

begin
  if NoAccess(H) then exit;

{  Erase BAM = set all sectors free  }
  with DriveHandles[H] do with Device[hbDevNo]^ do
   begin
    if FSType = EmuPC then exit;

    with Drv[hbDrvNo] do
     begin
      if FSType = EmuCBM then   {%%%%}
       begin
{
        hFileName := chCsep;
        ChangeDirectory(H);
        new(ADir);

        lADir := 1;
        DirTel := 0;
        GetDir(0, CurDir);
        ADir^[lADir] := CurDir;



        dispose(ADir);
        sDirPath := sDirPathOrg;
        hFileName := sDirPath;
        ChangeDirectory(H);
}
        ChDir(OrgDir);
        exit;
       end;

{ copy all BAMs in temporary buffers }
      for b1 := 0 to MaxBAMs(hbDevNo)-1 do
       begin
        ReadSector(H, BAMtrack(H, b1), BAMsector(H, b1));
        for b2 := 0 to 255 do tBAM[b1, b2] := dDataBlock[b2];
       end;

{ Do all needed changes here }
      for b1 := 1 to MaxTracks(H) do
       begin
        for b2 := 0 to MaxSectors(H, b1)-1 do
         begin
          SeekBAM;

          tBAM[bBAM, bByte] := tBAM[bBAM, bByte] or bBit;

          case Device[hbDevNo]^.FSType of
            Emu3040,
            Emu4040,
            Emu1541,
            Emu1571 : inc(tBAM[bBAM, b1 * 4]);
            Emu8250,
            Emu8050 : inc(tBAM[bBAM, 1 + b1 * 5]);
          end;
         end;

        SeekBAM;

        case Device[hbDevNo]^.FSType of
          Emu3040,
          Emu4040,
          Emu1541 : tBAM[bBAM, b1 * 4] := MaxSectors(H, b1);
          Emu1571 :
            if b1 < 36 then tBAM[bBAM, b1 * 4] := MaxSectors(H, b1)
             else tBAM[bBAM, $DD + b1] := MaxSectors(H, b1 - 35);
          Emu8250,
          Emu8050 : tBAM[bBAM, 1 + b1 * 5] := MaxSectors(H, b1);
         end;
       end;

{ restore all BAMs }
      for b1 := 0 to MaxBAMs(hbDevNo)-1 do
       begin
        for b2 := 0 to 255 do dDataBlock[b2] := tBAM[b1, b2];
        WriteSector(H, BAMtrack(H, b1), BAMsector(H, b1));
       end;

    {  Set BAM and info sector(s)  }
      SetSector(H, INFOtrack(H), 0);

      for b1 := 0 to MaxBAMs(hbDevNo)-1 do
       SetSector(H, BAMtrack(H, b1), BAMsector(H, b1));

      ValidateDir(H, DIRtrack(hbDevNo), DIRsector(hbDevNo));
     end;
   end;
end;


procedure ValidateDir(H, fTrack, fSector : byte);
{-----------------------------------------------------------------------------
This procedure checks all entries for valid files and then sets all records
belonging to a file.
This procedure is recursive in case of directories.
-----------------------------------------------------------------------------}
Var
  b1, fT, fS : byte;
  P : PDirEntryDxx;
  DirEntryDxx : TDirEntryDxx;
  wCnt : word;
begin
  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
  while (fTrack > 0) do
   if SectorIsFree(H, fTrack, fSector) then
    begin
     SetSector(H, fTrack, fSector);
     ReadSector(H, fTrack, fSector);

{  Check all valid files in this sector  }
     for b1 := 0 to 7 do
      begin
       P := PDirEntryDxx(Ptr(Seg(dDataBlock), Ofs(dDataBlock) + 2 + (b1 * $20)));
       Move(P^, DirEntryDxx, Sizeof(DirEntryDxx));
       wCnt := 0;

       if (DirEntryDxx.Filetype and $80 > 0) then
        begin                           { if valid then ..... }
         fT := DirEntryDxx.Track_1;
         fS := DirEntryDxx.Sector_1;

         if (DirEntryDxx.Filetype and $86 = $86) then
          begin
           ValidateDir(H, fT, fS);      { validate subdirectory }

{  Count all records of this subdirectory  }
           wCnt := 0;
           while (fT > 0) do            { Follow the chain of this file }
            begin
             inc(wCnt);
             ReadSector(H, fT, fS);
             fT := dDataBlock[0];
             fS := dDataBlock[1];
            end;
          end
         else
          begin
           while (fT > 0) do            { Follow the chain of this file }
            if SectorIsFree(H, fT, fS) then
             begin
              inc(wCnt);
              SetSector(H, fT, fS);
              ReadSector(H, fT, fS);
              fT := dDataBlock[0];
              fS := dDataBlock[1];
             end
            else fT := 0;               { STOP, circular reference !!! }

{  Side sectors of relative files  }
           if (DirEntryDxx.Filetype and $87 = $84) then
            begin
             fT := DirEntryDxx.TSideSec;
             fS := DirEntryDxx.SSideSec;

             while (fT > 0) do
              if SectorIsFree(H, fT, fS) then
               begin
                inc(wCnt);
                SetSector(H, fT, fS);
                ReadSector(H, fT, fS);
                fT := dDataBlock[0];
                fS := dDataBlock[1];
               end
              else fT := 0;             { STOP, circular reference !!! }
            end;
          end;

         ReadSector(H, fTrack, fSector);{ Restore original dDataBlock }
        end
         else DirEntryDxx.Filetype := 0;{ scratch any * type }

       DirEntryDxx.Lo_Size := lo(wCnt);
       DirEntryDxx.Hi_Size := hi(wCnt);
       Move(DirEntryDxx, P^, Sizeof(DirEntryDxx));
       WriteSector(H, ftrack, fsector); { write possible changes to disk }
      end;

     fTrack := dDataBlock[0];
     fSector := dDataBlock[1];
    end
   else fTrack := 0;                    { STOP, circular reference !!! }
end;


procedure WriteBlockDxx(H : Integer; WriteData : Word;
                               Var Data; Var DataWritten : Word);
Var
  WD : Byte;
  P : Pointer;
begin
  P           := @Data;
  DataWritten := 0;

  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
  while WriteData > 0 do
   begin

{  First read it  }
    ReadSector(H, hTrack, hSector);

    if hOffset > 255 then
     begin
      if dDataBlock[0] = 0 then
       begin
        AddSector(H);
        if DiskStatus <> 0 then exit;
       end
      else
       begin
{*  This is easy, we're overwriting so let continue to the next sector  *}
        hTrack  := dDataBlock[0];
        hSector := dDataBlock[1];
       end;

      hOffset := 2;
      Fillchar(dDataBlock, Sizeof(dDataBlock), 0);
     end;

    if WriteData > 256-hOffset then WD := 256-hOffset else WD := WriteData;

{*  Update vars  *}
    WriteData   := WriteData - WD;
    DataWritten := DataWritten + WD;

{*  Move the data  *}
    Move(P^, dDataBlock[hOffset], WD);
    P := Pointer(Ptr(Seg(P^), Ofs(P^)+WD));

{*  Update internals  *}
    hOffset := hOffset + WD;
    hPosition := hPosition + WD;

{*  Last block  *}
    dDataBlock[0] := 0;
    dDataBlock[1] := hOffset-1;

{*  Write it back  *}
    WriteSector(H, hTrack, hSector);
   end;
end;


procedure WriteSector(H, T, S : byte);
{----------------------------------------------------------------------------
Writes sector to disk
in:     T
        S
----------------------------------------------------------------------------}
var
  DxxFile : File;
  l : longint;
begin
  l := Translate2AbsPos(H, T, S);
  if (l = -1) then exit;

  with DriveHandles[H] do
   Assign(DxxFile, Device[hbDevNo]^.Drv[hbDrvNo].PCfile);
  Reset(DxxFile, 1);
  Seek(DxxFile, l);

  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
  BlockWrite(DxxFile, dDataBlock, Sizeof(dDataBlock));

  Close(DxxFile);

  with DriveHandles[H] do
   with Device[hbDevNo]^.Drv[hbDrvNo] do
    begin
     LastTrack := T;
     LastSector := S;
    end;
end;


begin
  boTest := True;
  boTestEnd := False;
  bByteInCnt1 := 0;
  bByteInCnt2 := 1;
end.







